
/**
    Build index
    (C) Copyright Magnus Galåen (magnusga@stud.ntnu.no)
 */

#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>

#include <fstream>
#include <iostream>
#include <list>
#include <set>
#include <map>
#include <string>
#include <vector>

#include "common/utf8-strings.h"
#include "parser/html_parser.h"

//#define VERIFY(a)	if ((a)==0) { cerr << "Runtime error at line " << __LINE__ << endl; perror("fwrite"); exit(-1); }
#define VERIFY(a)	(a)

using namespace std;



map< string,list<int> >	D;
vector<string>	W;
int		id = 0;



void fn( char *word, int pos, enum parsed_unit pu, enum parsed_unit_flag puf, void *wordlist )
{
    if (pu == pu_word)
	{
	    char	*w_lower = strdup(word);
	    convert_to_lowercase((unsigned char*)w_lower);
	    string	w(w_lower);

	    W.push_back(w);
	    map< string,list<int> >::iterator	it = D.find(w);

	    if (it==D.end())
		{
		    list<int>	L;
		    L.push_back(id);
		    D.insert( pair< string,list<int> >(w,L) );
		}
	    else
		{
		    int		last = *(--((*it).second.end()));
		    if (last != id)
			(*it).second.push_back(id);
		}

	    free(w_lower);
	}
}



int main(int argc, const char* argv[])
{
    if (argc!=1)
	{
//	    cout << "Usage: cat list_of_files.txt | " << argv[0] << " <dictionary> <ii> <id> <summary>" << endl;
	    cout << "Usage: cat list_of_files.txt | " << argv[0] << endl;
	    exit(-1);
	}

    FILE	*f_dictionary = fopen("dictionary","w"),
		*f_ii = fopen("ii","w"),
		*f_id = fopen("id","w");

    if (!f_dictionary || !f_ii || !f_id)
	{
	    cout << "Could not open files for writing!" << endl;
	    exit(-1);
	}

    html_parser_init();

    string	url, fname, category, number;
    int		pos = 0;

    ofstream	f_docs("matrix_docs");

    cout << "Reading files:" << endl;
    while (cin >> url >> fname >> category >> number)
	{
	    string	ifname = "files/" + fname;
    	    FILE	*fraw_html = fopen(ifname.c_str(), "r");
//	    if (id%10 == 0) { cout << "."; cout.flush(); }
//	    cout << "[" << fname << "]"; cout.flush();

	    f_docs << id << ": " << id << endl;

	    cout << "[" << fname << "]" << endl;
	    if (fraw_html==NULL)
        	{
        	    cerr << "Error opening " << ifname << endl;
		    continue;
        	}

    	    struct stat fileinfo;
	    fstat( fileno(fraw_html), &fileinfo );

    	    int         size = fileinfo.st_size;
	    char        *buf = (char*)malloc(size);

//	    cout << "[read]"; cout.flush();
    	    for (int i=0; i<size; )
    		{
        	    i+= fread( (void*)&(buf[i]), sizeof(char), size-i, fraw_html );
        	}

    	    char		*title, *body;

//	    cout << "[parse]"; cout.flush();
	    char		*c_url = strdup(url.c_str());
	    html_parser_run( c_url, buf, size, &title, &body, fn, NULL );
	    free(c_url);

	    free(buf);
	    fclose(fraw_html);

//	    cout << "[write(summary)]"; cout.flush();
	    int		body_len = strlen(body);
//	    VERIFY( fwrite( &body_len, sizeof(int), 1, f_summary ) );
	    string	f_summary_name = "preparsed/";
	    f_summary_name+= fname;
	    FILE	*f_summary = fopen(f_summary_name.c_str(), "w");
	    if (!f_summary)
		{
		    cout << "Could not open file for writing: " << f_summary_name << endl;
		    exit(-1);
		}
	    VERIFY( fwrite( body, body_len, 1, f_summary ) );
//	    pos+= sizeof(int) + body_len;
	    fclose(f_summary);

//	    cout << "[write(wordlist)]"; cout.flush();
	    string	f_wordlist_name = "wordlist/";
	    f_wordlist_name+= fname;
	    ofstream	of_wordlist(f_wordlist_name.c_str());
	    if (!of_wordlist)
		{
		    cout << "Could not open file for writing: " << f_summary_name << endl;
		    exit(-1);
		}
	    for (int i=0; i<W.size(); i++)
		of_wordlist << W[i] << endl;
	    W.clear();
	    of_wordlist.close();

//	    cout << "[write(pos,title)]"; cout.flush();
	    char		p_url[128], p_path[64], p_cat[64], p_title[124];
	    memset( p_url, '\0', 128 );
	    strncpy( p_url, url.c_str(), 127 );
	    memset( p_path, '\0', 64 );
	    strncpy( p_path, fname.c_str(), 63 );
	    memset( p_cat, '\0', 64 );
	    strncpy( p_cat, category.c_str(), 63 );
	    memset( p_title, '\0', 124 );
	    strncpy( p_title, title, 123 );

//	    VERIFY( fwrite( &pos, sizeof(int), 1, f_id ) );
	    VERIFY( fwrite( p_title, 124, 1, f_id ) );
	    VERIFY( fwrite( p_url, 128, 1, f_id ) );
	    VERIFY( fwrite( p_path, 64, 1, f_id ) );
	    VERIFY( fwrite( p_cat, 64, 1, f_id ) );

	    id++;
//	    cout << endl;

	    free(title);
	    free(body);
	}

    html_parser_exit();
    fclose(f_id);
    f_docs.close();

    cout << endl;
    cout << "Writing dictionary: "; cout.flush();
    int		nr = 0;

    ofstream	f_words("matrix_words");
//    set<int>	*dw = new set<int>[id];
    // document, list_of_words

    f_words << D.size() << endl;

    pos = 0;
    map< string,list<int> >::iterator	it = D.begin(), stop = D.end();
    for (; it!=stop; ++it,nr++)
	{
	    if (nr % 1000 == 0) { cout << "."; cout.flush(); }
	    char	word[28];
	    memset( word, '\0', 28 );
	    strncpy( word, (*it).first.c_str(), 27 );
	    f_words << (*it).first << endl;

	    VERIFY( fwrite( word, 28, 1, f_dictionary ) );
	    VERIFY( fwrite( &pos, sizeof(int), 1, f_dictionary ) );
//	    cout << "pos: " << pos << endl;

	    list<int>::iterator	ii_it=(*it).second.begin(), ii_stop=(*it).second.end();
	    int		ii_size = (*it).second.size();
	    VERIFY( fwrite( &ii_size, sizeof(int), 1, f_ii ) );
	    pos++;
	    for (; ii_it!=ii_stop; ++ii_it)
		{
		    int		_id = *ii_it;

//		    dw[_id].insert(nr);
		    VERIFY( fwrite( &_id, sizeof(int), 1, f_ii ) );
		    pos++;
		}
	}

    fclose(f_dictionary);
    fclose(f_ii);
    cout << endl;

    // Print out sparse matrix (CCS):
    cout << "Generating sparse matrix: "; cout.flush();

    // matrix_words, second run:
    nr = 0;
    it = D.begin(), stop = D.end();
    for (; it!=stop; ++it,nr++)
	{
	    f_words << nr << " " << (*it).second.size() << " " << (*it).second.size() << endl;
	}

    f_words.close();

    ofstream	f_col_ccs("matrix_col_ccs"), f_row_ccs("matrix_row_ccs"), f_tfn_nz("matrix_tfn_nz");
    int		rows, cols, nz;
    int		rowptr=0, colptr=0;

    rows = D.size();
    cols = id;
    nz = 0;

    /***/
    set<int>	*dw = new set<int>[id];
    // document, list_of_words

    nr = 0;

    it = D.begin(), stop = D.end();
    for (; it!=stop; ++it,nr++)
	{
	    if (nr % 1000 == 0) { cout << "."; cout.flush(); }

	    list<int>::iterator	ii_it=(*it).second.begin(), ii_stop=(*it).second.end();
	    for (; ii_it!=ii_stop; ++ii_it)
		{
		    int		_id = *ii_it;
		    dw[_id].insert(nr);
		}

	    (*it).second.clear();
	}
    /***/

    f_col_ccs << "0" << endl;

    for (int i=0; i<id; i++)
	{
	    set<int>::iterator	s_it=dw[i].begin(), s_stop=dw[i].end();
	    for (; s_it!=s_stop; ++s_it)
		{
		    f_row_ccs << *s_it << endl;
		    f_tfn_nz << "1.0" << endl;
		    colptr++;
		}

	    f_col_ccs << colptr << endl;
	}

    nz = colptr;

    f_col_ccs.close();
    f_row_ccs.close();
    f_tfn_nz.close();

    ofstream	f_dim("matrix_dim");

    f_dim << rows << " " << cols << " " << nz << endl;
    f_dim.close();
    cout << "done." << endl;
}


