
#include <fstream>
#include <string>
#include <set>
#include <map>

using namespace std;



class Dictionary
{
    public:

	Dictionary(char *dictionary)
	    {
		ifstream	idict(dictionary);
		string		word;

		while (idict >> word)
		    {
			for (int i=0; i<word.length(); i++)
			    if (word[i]=='_') word[i] = ' ';

			D.insert(word);
		    }
	    }

	~Dictionary()
	    {
	    }

	bool exist(string s)
	    {
		return D.find(s)!=D.end();
	    }

    private:

	set<string>	D;
};



class Stemmer
{
    public:

	Stemmer(Dictionary *_D, char *exception_list)
	: D(_D)
	    {
		ifstream	ilemma(exception_list);
		string		tense, stem;

		while (ilemma >> tense >> stem)
		    {
//			cout << "{" << tense << "," << stem << "}" << endl;
			for (int i=0; i<tense.length(); i++)
			    if (tense[i]=='_') tense[i] = ' ';

			for (int i=0; i<stem.length(); i++)
			    if (stem[i]=='_') stem[i] = ' ';

			lemma.insert( pair<string,string>(tense, stem) );
		    }
	    }

	~Stemmer()
	    {
	    }

	string stem(string s)
	    {
		if (s.length()==0) return s;

		map<string,string>::iterator	it=lemma.find(s);

		if (it!=lemma.end())
		    {
			return (*it).second;
		    }

		if (s.length() <= 2)
		    {
			return s;
		    }
//		if (s.length() <= 2 || is_suffix(s, "ss")) ...

		static string		sufx[20] = {
		    /* Noun suffixes */
		    "s", "ses", "xes", "zes", "ches", "shes", "men", "ies",
		    /* Verb suffixes */
		    "s", "ies", "es", "es", "ed", "ed", "ing", "ing",
		    /* Adjective suffixes */
		    "er", "est", "er", "est"
		};
		static string		addr[20] = {
		    /* Noun endings */
		    "", "s", "x", "z", "ch", "sh", "man", "y",
		    /* Verb endings */
		    "", "y", "e", "", "e", "", "e", "",
		    /* Adjective endings */
		    "", "", "e", "e"
		};

		string	stem, suffix;

		// Assume word is a noun:
		if (is_suffix(s, "ful"))
		    {
			stem.assign(s, 0, s.length()-3);
			suffix = "ful";
		    }
		else
		    {
			stem = s;
			suffix = "";
		    }

		for (int i=0; i<20; i++)
		    {
			if (is_suffix(stem, sufx[i]))
			    {
				string	ret;
				ret.assign(stem, 0, stem.length() - sufx[i].length());
				ret+= addr[i];

//				if (is_defined((char*)ret.c_str(),1) || is_defined((char*)ret.c_str(),2) || is_defined((char*)ret.c_str(),3) || is_defined((char*)ret.c_str(),4))
				if (D->exist(ret))
				    {
					ret+= suffix;
					return ret;
				    }
			    }
		    }

		return s;
	    }


    private:

	Dictionary		*D;
	map<string,string>	lemma;

	static inline bool is_suffix(string s1, string s2)
	    {
		int	s1_len = s1.length(), s2_len = s2.length();

		if (s2_len >= s1_len) return false;

		for (int i=0; i<s2_len; i++)
		    if (s1[s1_len-1-i] != s2[s2_len-1-i])
			return false;

		return true;
	    }
};

