
/**
 *	Calculate Entropy, (C) Copyright Magnus Galåen
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <math.h>

#include <iostream>
#include <fstream>
#include <list>
#include <map>
#include <set>

extern "C"
{
#include "query/query_parser.h"
}

#include "blocks.h"

using namespace std;


unsigned long long binomial(int n, int m)
{
    unsigned long long	**bc;//[n+1][n+1];
    unsigned long long	result;

    bc = new unsigned long long *[n+1];
    for (int i=0; i<=n; i++)
	bc[i] = new unsigned long long[n+1];

    for (int i=0; i<=n; i++) bc[i][0] = 1;

    for (int j=0; j<=n; j++) bc[j][j] = 1;


    for (int i=1; i<=n; i++)
	for (int j=1; j<i; j++)
	    bc[i][j] = bc[i-1][j-1] + bc[i-1][j];

    result = bc[n][m];

    for (int i=0; i<=n; i++)
	delete bc[i];
    delete bc;

    return result;
}



int main( int argc, const char *argv[] )
{
    if (argc<2)
	{
	    cout << "Usage: " << argv[0] << " <matrix_tfn_doctoclus.X> [<fasit.X>]" << endl;
	    exit(-1);
	}

    FILE	*f_id = fopen("id", "r");
    int		size_id;
    void	*mmap_id;
    struct stat	inode;

    fstat(fileno(f_id), &inode);
    size_id = inode.st_size;

    mmap_id = mmap(0, size_id, PROT_READ, MAP_PRIVATE, fileno(f_id), 0);

    const char	*ptr = &argv[1][strlen(argv[1])-1];
    while (*ptr!='.') ptr--;
    ptr++;

    int		num_clusters = atoi(ptr), num_documents = 0;
    cout << num_clusters << " clusters." << endl;
    list<int>	C[num_clusters];
    ifstream	f_clusters(argv[1]);
    int		cnr, docid;

    bool	fasit=false;
    set< pair<int,int> >	F;
    int		f_max=0;

    if (argc==3)
	{
	    fasit = true;
	    ifstream	_ff(argv[2]);
	    int		c, d;
	    while (_ff >> c >> d)
		{
		    F.insert( pair<int,int>(c,d) );
		    if (c>f_max) f_max = c;
		}

	    f_max++;
	}


    set<string>	unique_categories;

    f_clusters >> docid;

    while (f_clusters >> cnr >> docid)
	{
	    C[cnr].push_back(docid);
	    num_documents++;

	    struct id_block	*doc = (struct id_block*)mmap_id;
	    doc+= docid;

	    unique_categories.insert(doc->category);
	}


    double	entropy_first_sum = 0.0,
		entropy_second_sum = 0.0;

    if (!fasit)
    {
    for (int c=0; c<num_clusters; c++)
	{
	    list<int>::iterator	it=C[c].begin(), stop=C[c].end();
	    map<string,int>	Category;

	    for (; it!=stop; ++it)
		{
		    struct id_block	*doc = (struct id_block*)mmap_id;
		    doc+= *it;

		    map<string,int>::iterator	m_it = Category.find(doc->category);
		    if (m_it==Category.end())
			{
			    Category.insert( pair<string,int>(doc->category, 1) );
			}
		    else
			{
			    (*m_it).second++;
			}
		}

	    map<string,int>::iterator	m_it=Category.begin(), m_stop=Category.end();
	    double			h_k = C[c].size();
	    for (; m_it!=m_stop; ++m_it)
		{
		    double	h_ck = (*m_it).second;
		    entropy_first_sum+= (h_ck/(double)num_documents) * log(h_ck/h_k);
		}

	    entropy_second_sum+= log((double)binomial((int)h_k + unique_categories.size() -1, unique_categories.size() -1));
	}
    }
    else
    {
    for (int c=0; c<num_clusters; c++)
	{
	    list<int>::iterator	it=C[c].begin(), stop=C[c].end();
	    double		h_ck[f_max];

	    for (int k=0; k<f_max; k++)
		h_ck[k] = 0;

	    for (; it!=stop; ++it)
		{
		    for (int k=0; k<f_max; k++)
			{
			    set< pair<int,int> >::iterator	s_it=F.find(pair<int,int>(k,*it));

			    if (s_it!=F.end()) h_ck[k]++;
			}
		}

	    double			h_k = C[c].size();
	    for (int k=0; k<f_max; k++)
		{
//		    cout << "h_k = " << h_k << endl;
//		    cout << "h_ck[" << k << "] = " << h_ck[k] << endl;
		    if (h_ck[k]>0) entropy_first_sum+= (h_ck[k]/(double)num_documents) * log(h_ck[k]/h_k);
		}

	    entropy_second_sum+= log((double)binomial((int)h_k + f_max -1, f_max -1));
//	    cout << "log( binomial( "
//	    << ((int)h_k + f_max -1) << "," << (f_max -1) << " ) ) = log( "
//	    << (double)binomial((int)h_k + f_max -1, f_max -1) << " ) = "
//	    << log((double)binomial((int)h_k + f_max -1, f_max -1)) << endl;
	}
    }

    double	entropy = - entropy_first_sum + (1/(double)num_documents)*entropy_second_sum;
//    cout << "first: " << entropy_first_sum << endl;
//    cout << "second: " << entropy_second_sum << endl;
    cout << "Entropy: " << entropy << endl;
}

