

/**
    (C) Copyright Magnus Galåen (magnusga@idi.ntnu.no)
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <fstream>
#include <iostream>
#include <string>
#include <vector>
#include <set>

#include "dictionary.cpp"
#include "blocks.h"

extern "C"
{
#include "common/utf8-strings.h"
#include "parser/html_parser.h"
#include "query/query_parser.h"
#include "generateSnippet/snippet.parser.h"
}

using namespace std;


Dictionary	*dictionary;
Stemmer		*stemmer;
int		size_id;
void		*mmap_id;


void io_init()
{
    dictionary = new Dictionary("dict/dictionary.english");
    stemmer = new Stemmer(dictionary, "dict/lemma.english");

    html_parser_init();

    FILE	*f_id = fopen("id", "r");
    struct stat	inode;

    fstat(fileno(f_id), &inode);
    size_id = inode.st_size;

    mmap_id = mmap(0, size_id, PROT_READ, MAP_PRIVATE, fileno(f_id), 0);
}



struct fn_data
{
    bool		filter_washing, filter_stemming, paragraphs_only;
    vector<string>	*Doc, *Snippet;
    set<string>		*SW;
    map<string,int>	*nW;
    int			s, t;
};




void fn_full( char *word, int pos, enum parsed_unit pu, enum parsed_unit_flag puf, void *wordlist )
{
    struct fn_data	*data = (struct fn_data*)wordlist;
/*
    switch (pu)
	{
	    case pu_word:
		    if (word_is_defined(word))
			{
		    	    cout << " {" << word << "}";
			    data->s++;
			}
		    else
			{
			    cout << " " << word;
			    data->t++;
			}
		    break;
	    case pu_linkword: cout << " [" << word << "]"; break;
	    case pu_eos: cout << "." << endl; break; // "(" << data->s << "/" << (data->s+data->t) << ")" << endl; data->s = data->t = 0; break;
	    case pu_paragraph:
		    if (data->s >= 40 && (double)data->s/(double)(data->t + data->s) >= 0.6)
			cout << "PARAGRAPH";
		    else
			cout << "GARBAGE";
		    cout << " [" << data->s << "/" << (data->t + data->s) << " = " << (double)data->s/(double)(data->t + data->s) << "]" << endl;
		    data->s = data->t = 0;
		    cout << "-----" << endl;
		    break;
	}
*/

    if (pu == pu_word || pu == pu_linkword)
	{
	    char		*w_lower = strdup(word);
	    convert_to_lowercase((unsigned char*)w_lower);
	    string		w(w_lower);

	    bool	add = true;

	    if (data->filter_washing)
		{
		    if (w.size() <= 2) add = false;
		    if (w[0]>='0' && w[0]<='9') add = false;
		}

	    if (add && data->filter_stemming)
		{
		    w = stemmer->stem(w);
		}

	    if (add && data->SW!=NULL)
		{
		    set<string>::iterator	it = data->SW->find(w);
		    if (it!=data->SW->end()) add = false;
		}

	    if (add)
		{
//		    vector<string>	&V = *(vector<string>*)wordlist;
/*
		    char	s[1024];
		    if (pu==pu_word && word_is_defined((char*)w.c_str()))
			{
			    data->Doc->push_back(w);
			}
		    else if (pu==pu_linkword)
			{
			    sprintf(s, "[%s]", w.c_str());
			    data->Doc->push_back(string(s));
			}
		    else
			{
			    sprintf(s, "(%s)", w.c_str());
			    data->Doc->push_back(string(s));
			}
*/
		    data->Doc->push_back(w);
		    map<string,int>::iterator	it = data->nW->find(w);
		    if (it == data->nW->end())
		        data->nW->insert( pair<string,int>(w,1) );
		    else
			(*it).second++;

		    if (data->paragraphs_only)
			{
			    if (pu==pu_word && dictionary->exist(w)) data->s++;
			    data->t++;
			}
		}

	    free(w_lower);
	}
    else if (pu == pu_paragraph && data->paragraphs_only)
	{
	    // If data since last paragraph is garbage, remove it:
	    if (!(data->s >= 40 && (double)data->s/(double)data->t >= 0.6))
		{
		    for (int j=0; j<data->t; j++)
			{
			    map<string,int>::iterator	it = data->nW->find((*data->Doc)[data->Doc->size()-1]);
			    (*it).second--;
			    data->Doc->pop_back();
			}
		}
//		    char	s[1024];
//		    sprintf(s, "\nGARBAGE [%i/%i = %f]\n\n", data->s, data->t, (double)data->s/(double)data->t);
//		    data->Doc->push_back(string(s));
//		}
//	    else
//		{
//		    char	s[1024];
//		    sprintf(s, "\nPARAGRAPH [%i/%i = %f]\n\n", data->s, data->t, (double)data->s/(double)data->t);
//		    data->Doc->push_back(string(s));
//		}

	    data->s = data->t = 0;
	}
}



class Document
{
    public:

	Document(int _docid)
	: docid(_docid), body(NULL), snippet(NULL)
	    {
		data.filter_washing = true;
		data.filter_stemming = false;
		data.paragraphs_only = false;
		data.Doc = NULL;
		data.SW = NULL;
		data.nW = NULL;
		data.Snippet = NULL;

		// Get document info:
	        struct id_block	*doc_d = (struct id_block*)mmap_id;
	        doc_d+= docid;

		url = doc_d->url;
		path = doc_d->path;
		category = doc_d->category;
		title = doc_d->title;
		for (int i=0; title[i]!='\0'; i++)
		    if (title[i]<32) title[i] = ' ';

		// Read raw html:
	        string	filename = "files/";
	        filename+= doc_d->path;

	        FILE        *file = fopen((char*)filename.c_str(), "r");

	        if (!file)
	            {
	                cerr << "input.cpp: Error! Could not open " << filename << endl;
			rawHTML = NULL;
	                return;
	            }

	        // Get filesize:
	        struct stat fileinfo;
	        fstat( fileno( file ), &fileinfo );

	        rawHTMLsize = fileinfo.st_size;
	        rawHTML = new char[rawHTMLsize];

	        for (int i=0; i<rawHTMLsize;)
	            {
	                i+= fread( (void*)&(rawHTML[i]), sizeof(char), rawHTMLsize-i, file );
	            }

	        fclose(file);
	    }

	~Document()
	    {
		free(rawHTML);
		if (body!=NULL) free(body);
		if (snippet!=NULL) free(snippet);
		if (data.Doc!=NULL) delete data.Doc;
		if (data.SW!=NULL) delete data.SW;
		if (data.nW!=NULL) delete data.nW;
		if (data.Snippet!=NULL) delete data.Snippet;
	    }

	int getDocid() { return docid; }
	string getUrl() { return url; }
	string getPath() { return path; }
	string getCategory() { return category; }
	string getTitle() { return title; }
	char* getSnippet() { return snippet; }

	vector<string>* getDocData() { return data.Doc; }
	vector<string>* getSnippetData() { return data.Snippet; }
	int getWordOccurences( string s )
	    {
		map<string,int>::iterator	it = data.nW->find(s);
		if (it== data.nW->end()) return 0;
		return (*it).second;
	    }

	void setFilterWashing(bool state)
	    {
		data.filter_washing = state;
	    }

	void setFilterStemming(bool state)
	    {
		data.filter_stemming = state;
	    }

	void setFilterStopWords( set<string> *_SW )
	    {
		data.SW = _SW;
	    }

	vector<string>* getdataFullDocument()
	    {
		// If file couldn't be read:
		if (rawHTML==NULL)
		    {
			cerr << "input.cpp: Warning! " << __LINE__ << endl;
			return NULL;
		    }

		char		*temp_title;

		if (data.Doc != NULL) delete data.Doc;
		if (data.nW != NULL) delete data.nW;
		data.Doc = new vector<string>();
		data.nW = new map<string,int>();

//		cout << endl << "HTML:";
		html_parser_run( (char*)url.c_str(), rawHTML, rawHTMLsize, &temp_title, &body, fn_full, (void*)&data );
//		cout << endl;
//		exit(-1);

		body_size = strlen(body);
		free(temp_title); // Title has already been stored.

		return data.Doc;
	    }

	vector<string>* getdataParagraphsOnly()
	    {
		// If file couldn't be read:
		if (rawHTML==NULL)
		    {
			cerr << "input.cpp: Warning! " << __LINE__ << endl;
			return NULL;
		    }

		char		*temp_title;

		if (data.Doc != NULL) delete data.Doc;
		if (data.nW != NULL) delete data.nW;
		data.Doc = new vector<string>();
		data.nW = new map<string,int>();

		data.paragraphs_only = true;
		data.s = 0;
		data.t = 0;

		html_parser_run( (char*)url.c_str(), rawHTML, rawHTMLsize, &temp_title, &body, fn_full, (void*)&data );

		data.paragraphs_only = false;

		body_size = strlen(body);
		free(temp_title); // Title has already been stored.

		return data.Doc;
	    }

	vector<string>* getdataRelevantParagraphs()
	    {
	    }

	// getdataFullDocument() must be run first!
	vector<string>* getdataSnippet( query_array *qa, int snippet_size, bool eat=true )
	    {
		if (snippet_size > 960)
		    {
			cerr << "input.cpp: Warning! Snippet_size too large." << endl;
			return NULL;
		    }

		if (data.Doc->empty() || body[0]=='\0')
		    {
			if (snippet!=NULL) free(snippet);
			snippet = strdup("");
			return new vector<string>();
		    }

		generate_snippet(*qa, body, body_size, &snippet, "<b>", "</b>", snippet_size);

		for (int i=0; snippet[i]!='\0'; i++)
		    if (snippet[i]<32) snippet[i] = ' ';

		if (!eat) return NULL;

		if (data.Snippet!=NULL) delete data.Snippet;
		data.Snippet = new vector<string>();
		if (data.nW!=NULL) delete data.nW;
		data.nW = new map<string,int>();

		for (int k=0, p=0; snippet[k]!='\0' && k<snippet_size+40; k++)
		    {
			while (snippet[k]!='\0' && !(isalpha(snippet[k]) || isdigit(snippet[k]))) k++;
		        p = k;
			while (snippet[p]!='\0' && (isalpha(snippet[p]) || isdigit(snippet[p]) || snippet[p]=='\'')) p++;
		        while (snippet[p-1]=='\'') p--;

		        if (p>k+1)
		    	    {
		    		char	buf[1024];
			        strncpy(buf, &(snippet[k]), p-k);
				buf[p-k] = '\0';
			        convert_to_lowercase((unsigned char*)buf);
				string	w(buf);
				bool	add = true;

		    	        if (data.filter_washing)
				    {
					if (w.size() <= 2) add = false;
					if (w[0]>='0' && w[0]<='9') add = false;
				    }

				if (add && data.filter_stemming)
				    {
					w = stemmer->stem(w);
				    }

			        if (add && data.SW!=NULL)
				    {
				        set<string>::iterator	it = data.SW->find(w);
					if (it!=data.SW->end()) add = false;
				    }


				if (add)
			    	    {
					data.Snippet->push_back(w);

					map<string,int>::iterator	it = data.nW->find(w);
					if (it == data.nW->end())
					    data.nW->insert( pair<string,int>(w,1) );
					else
					    (*it).second++;
				    }
			    }

			k = p;
		    }

		return data.Snippet;
	    }


    private:

	int		docid;
	string		url, path, category, title;
	char		*rawHTML, *body, *snippet;
	int		rawHTMLsize, body_size;
	struct fn_data	data;
};

