#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <math.h>

#include <iostream>
#include <vector>
#include <map>
#include <set>
#include <string>

#include "common/utf8-strings.h"
#include "blocks.h"
#include "dictionary.cpp"

using namespace std;


struct idf_node
{
    char	w[28];
    float	f;
};


int main(int argc, char *argv[])
{
    Dictionary	*dictionary;
    Stemmer	*stemmer;

    dictionary = new Dictionary("dict/dictionary.english");
    stemmer = new Stemmer(dictionary, "dict/lemma.english");


    FILE	*f_dictionary = fopen("dictionary", "r"),
		*f_ii = fopen("ii", "r"),
		*f_id = fopen("id", "r");
    int		size_dictionary, size_ii, size_id;
    int		n_dictionary, n_docs;
    void	*mmap_dictionary, *mmap_ii, *mmap_id;
    struct stat	inode;

    fstat(fileno(f_dictionary), &inode);
    size_dictionary = inode.st_size;
    n_dictionary = size_dictionary / 32;
    fstat(fileno(f_ii), &inode);
    size_ii = inode.st_size;
    fstat(fileno(f_id), &inode);
    size_id = inode.st_size;
    n_docs = size_id / sizeof(struct id_block);

    mmap_dictionary = mmap(0, size_dictionary, PROT_READ, MAP_PRIVATE, fileno(f_dictionary), 0);
    mmap_ii = mmap(0, size_ii, PROT_READ, MAP_PRIVATE, fileno(f_ii), 0);
    mmap_id = mmap(0, size_id, PROT_READ, MAP_PRIVATE, fileno(f_id), 0);

    if (mmap_dictionary == MAP_FAILED || mmap_ii == MAP_FAILED || mmap_id == MAP_FAILED)
	{
	    perror("mmap");
	    exit(-1);
	}

    cout << "Generating idf-list: "; cout.flush();
    map<string, set<int> >	IDF;

    struct dict_block	*ptr_dict = (struct dict_block*)mmap_dictionary;
    for (int i=0; i<n_dictionary; i++)
	{
	    if (i%50000 == 0) { cout << "."; cout.flush(); }

	    int		*id_list = (int*)mmap_ii;
	    id_list+= ptr_dict->addr;
	    int		num_word_docs = *id_list;

	    char	*w_lower = strdup(ptr_dict->word);
	    convert_to_lowercase((unsigned char*)w_lower);
	    string	w(w_lower);
	    free(w_lower);
	    if (w.length() <= 2 || w[0]<'a' || w[w.size()-1]<'a')
		{
		    ptr_dict++;
		    continue;
		}
	    w = stemmer->stem(w);

	    set<int>	Doclist;
	    for (int j=0; j<num_word_docs; j++)
		{
		    ++id_list;
		    Doclist.insert(*id_list);
		}

	    map<string, set<int> >::iterator	it = IDF.find(w);
	    if (it!=IDF.end())
		{
		    (*it).second.insert(Doclist.begin(), Doclist.end());
		}
	    else
		{
		    IDF.insert( pair<string, set<int> >(w, Doclist) );
		}

	    ptr_dict++;
	}

    cout << "done" << endl;

    FILE	*f_idf = fopen("idf", "w");
    struct idf_node	*node = (struct idf_node*)malloc(sizeof(struct idf_node));

    map<string, set<int> >::iterator	it=IDF.begin(), stop=IDF.end();
    for (;it!=stop; ++it)
	{
	    if ((*it).second.size() > 5)
		{
		    double	idf = log( (double)(1+n_docs) / (double)(*it).second.size() );
//		    cout << (*it).first << " " << idf << endl;
		    strncpy(node->w, (*it).first.c_str(), 28);
		    node->w[27] = '\0';
		    node->f = idf;
		    fwrite( node, sizeof(struct idf_node), 1, f_idf );
		}
	}
}
