
#ifndef _OUTPUT_H_
#define _OUTPUT_H_


#include <sys/mman.h>

#include <iostream>
#include <fstream>
#include <set>
#include <list>

using namespace std;


extern "C"
{
#include "query/query_parser.h"
#include "generateSnippet/snippet.parser.h"
}

#include "blocks.h"


class Output
{
    public:

	Output( char* filename, query_array *_qa, void *_mmap_id )
	    : bk(0), output_stream(filename), qa(_qa), mmap_id(_mmap_id)
	    {
//		output_stream.open(filename);

		if (!output_stream)
		    {
			cerr << "Error: Could not open " << filename << " for writing!" << endl;
			return;
		    }

		output_stream << "\
<?xml version=\"1.0\" encoding=\"utf-8\"?>\
<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\
<html xmlns=\"http://www.w3.org/1999/xhtml\">\
<head>\
    <title>Search: \"";

		for (int i=0; i<qa->n; i++)
		    output_stream << qa->query[i].s[0] << " ";

		output_stream << "\"</title>\
</head>\
<body style=\"font-size:smaller\">\
    <table style=\"padding:10px\">\
	<tr style=\"vertical-align: top\">\
	    <td style=\"width:200px;\"><div style=\"width:140px; border-style:solid; border-size:1; padding:16px; border-color:#6f9fbf; background:#afdfff; vertical-align:top\">" << endl;
	    }

	~Output()
	    {
		if (!output_stream) return;
		output_stream << "</td></tr></table>" << endl;
		output_stream << "</body></html>" << endl;
		output_stream.close();
	    }

	void print_sidebar( int num_documents )
	    {
		if (!output_stream) return;

		output_stream << "Found " << num_documents << " documents.\
	    </div></td>\
	    <td style=\"width:600px\">" << endl;
	    }

	void print_hr()
	    {
		if (!output_stream) return;

		output_stream << "<hr />" << endl;
	    }

#ifdef SIMPLE
	void print_snippets( list<int> L, bool verbose=true )
#else
	void print_snippets( list<int> L, set<string> W, bool verbose=true )
#endif
	    {
		if (!output_stream) return;

		if (bk==0)
		    {
			output_stream << "<div style=\"background: white; padding: 6px\">" << endl;
			bk = 1;
		    }
		else
		    {
			output_stream << "<div style=\"background: #dfe0d0; padding: 6px\">" << endl;
			bk = 0;
		    }

#ifndef SIMPLE
		output_stream << "<div style=\"color:gray\">" << endl;
		set<string>::iterator	w_it=W.begin(), w_stop=W.end();
		for (; w_it!=w_stop; ++w_it)
		    {
			if ((*w_it).find("tony")!=string::npos || (*w_it).find("blair")!=string::npos
			    || (*w_it).find("gordon")!=string::npos || (*w_it).find("brown")!=string::npos)
			    output_stream << "(" << (*w_it) << ") ";
		    }
		output_stream << "</div>" << endl;
#endif

		list<int>::iterator	it=L.begin(), stop=L.end();
		int			i;
		for (i=0; it!=stop && i<30; ++it,i++)
		    {
		        struct id_block	*doc = (struct id_block*)mmap_id;
		        doc+= *it;

			if (verbose)
		    	    cout << *it << " " << doc->url << " " << doc->path << " "
				<< doc->category << " \"" << doc->title << "\"" << endl;

			string	preparsed_fname = "preparsed/";
			preparsed_fname+= doc->path;
			FILE	*preparsed_f = fopen(preparsed_fname.c_str(), "r");

			output_stream << "<div>" << endl;
			output_stream << "<span style=\"color:blue\">" << doc->title << "</span><br />" << endl;
			output_stream << "<span style=\"color:gray; font-size:smaller\">DocID:" << *it << "</span><br />" << endl;

			if (!preparsed_f)
			    {
			        output_stream << "Error reading document.<br />" << endl;
			    }
			else
			    {
				struct stat	fileinfo;
				fstat( fileno(preparsed_f), &fileinfo );

				int		size = fileinfo.st_size;
				if (size>0)
				    {
					char	*buf = (char*)malloc(size);

					for (int j=0; j<size;)
					    j+= fread( &(buf[j]), sizeof(char), size-j, preparsed_f );

					char	*snippet;

					generate_snippet( *qa, buf, size, &snippet, "<strong>", "</strong>", 320 );

					if (snippet!=NULL)
					    {
						output_stream << snippet << "<br />" << endl;
						free(snippet);
					    }
					else
					    {
						output_stream << "<br />" << endl;
					    }

					free(buf);
				    }
				else
				    {
					output_stream << "<br />" << endl;
				    }

				fclose(preparsed_f);
			    }

			output_stream << "<span style=\"color:green; font-size:smaller\">" << doc->url << "</span><br />" << endl;
			output_stream << "</div><br />" << endl;
		    }

		if (i<L.size())
		    output_stream << "<div>" << (L.size() -i) << " documents were not shown.</div><br />" << endl;
		output_stream << "</div>" << endl;
	    }

    private:

	ofstream	output_stream;
	query_array	*qa;
	void		*mmap_id;
	int		bk;
};


#endif	// _OUTPUT_H_
