
#include <assert.h>
#include <limits.h>
#include <math.h>

#include <iostream>
#include <fstream>
#include <list>
#include <vector>
#include <set>
#include <map>
#include <string>
#include <sstream>

#include "blocks.h"
#include "input.cpp"

extern "C"
{
#include "query/query_parser.h"
}

using namespace std;


#define DUAL
#define	SNIPPET_LEN	320


struct idf_node
{
    char	w[28];
    float	f;
};

int cmp_idf(const void *a, const void *b)
{
    return strcmp((const char*)a, (const char*)b);
}



int main(int argc, char *argv[])
{
    if (argc<3)
	{
	    cerr << "Usage: " << argv[0] << " [<flags>] <result_docs> <query>" << endl;
	    cerr << "\t-snippet (default)" << endl;
	    cerr << "\t-paragraph" << endl;
	    cerr << "\t-means <prefix>" << endl;
	    exit(-1);
	}

    enum p_type { p_snippet, p_para }	process = p_snippet;
    bool	means = false;
    string	means_prefix;

    int		argnum = 1;
    while (argv[argnum][0]=='-')
	{
	    if (!strcmp(argv[argnum],"-snippet")) process = p_snippet;
	    else if (!strcmp(argv[argnum],"-paragraph")) process = p_para;
	    else if (!strcmp(argv[argnum],"-means"))
		{
		    means = true;
		    argnum++;
		    means_prefix = argv[argnum];
		}
	    argnum++;
	}

    ifstream	docs(argv[argnum]);

    if (!docs)
	{
	    cerr << "Error opening " << argv[argnum] << endl;
	    exit(-1);
	}


    // mmap inverted index:
    FILE	*f_idf = fopen("idf", "r");
    int		size_idf;
    int		n_idf;
    void	*mmap_idf;
    struct stat	inode;

    fstat(fileno(f_idf), &inode);
    size_idf = inode.st_size;
    n_idf = size_idf / sizeof(struct idf_node);

    mmap_idf = mmap(0, size_idf, PROT_READ, MAP_PRIVATE, fileno(f_idf), 0);

    if (mmap_idf == MAP_FAILED)
	{
	    perror("mmap");
	    exit(-1);
	}
    // end mmap

    query_array		qa;
    char		*query = argv[argnum+1];
    get_query(query, strlen(query), &qa);

    io_init();


    ifstream	sw_file("stopwords/ENG.txt");
    string	_stemp;
    set<string>	*stopwords = new set<string>();

    while (sw_file >> _stemp)
	{
	    stopwords->insert(_stemp);
	    stopwords->insert(stemmer->stem(_stemp));
	}

    // De eneste escapes som blir sendt med fra parseren:
    stopwords->insert("gt");
    stopwords->insert("lt");
    stopwords->insert("quot");
    stopwords->insert("amp");



    vector<int>		result;
    string		_dtemp;
    int			d;
    map<string,int>	S;
    vector<Document*>	VD;

    map< string, list<int> >	II;

    while (docs >> _dtemp >> d)
	{
	    result.push_back(d);
	}

    cout << result.size() << endl;

    ofstream	f_docs;
    if (means)
	{
	    string	f_docs_name = means_prefix + "_docs";
	    f_docs.open(f_docs_name.c_str());
	}

    for (int doc=0; doc<result.size(); doc++)
	{
	    Document	*D = new Document(result[doc]);

	    if (means) f_docs << result[doc] << ": " << result[doc] << endl;

	    cerr << result[doc] << endl;
	    cerr << D->getPath() << endl;

//	    cout << result[doc] << endl;
//	    cout << D->getUrl() << endl;
//	    cout << D->getPath() << endl;
//	    cout << D->getCategory() << endl;
//	    cout << D->getTitle() << endl;

	    D->setFilterWashing(true);
	    D->setFilterStemming(true);
//	    D->setFilterStopWords(stopwords);

	    vector<string>	*V;

	    if (process == p_snippet)
		{
		    D->getdataFullDocument();
		    V = D->getdataSnippet(&qa, SNIPPET_LEN);
		}
	    else if (process == p_para)
		{
		    V = D->getdataParagraphsOnly();
		    D->getdataSnippet(&qa, SNIPPET_LEN, false);
		}

//	    cout << D->getSnippet() << endl;

	    map<string,int>	L;
	    for (int i=0; V!=NULL && i<V->size(); i++)
		{
		    string	w = (*V)[i];
//		    cout << V[i] << " ";
		    L.insert(pair<string,int>(w,1));

		    map< string, list<int> >::iterator	it=II.find(w);
		    if (it!=II.end())
			{
			    (*it).second.push_back(doc);
			}
		    else
			{
			    list<int>	dl;
			    dl.push_back(doc);
			    II.insert( pair< string, list<int> >(w, dl) );
			}

#ifdef DUAL
		    if (i>0)
			{
			    string	w2 = (*V)[i-1] + " " + (*V)[i];

			    L.insert(pair<string,int>(w2,1));

			    map< string, list<int> >::iterator	it=II.find(w2);
			    if (it!=II.end())
				{
				    (*it).second.push_back(doc);
				}
			    else
				{
				    list<int>	dl;
				    dl.push_back(doc);
				    II.insert( pair< string, list<int> >(w2, dl) );
				}
			}
#endif
		}

	    map<string,int>::iterator	l_it=L.begin(), l_stop=L.end();
	    for (; l_it!=l_stop; ++l_it)
		{
		    map<string,int>::iterator	it=S.find((*l_it).first);
		    if (it == S.end())
			{
			    S.insert(pair<string,int>((*l_it).first,1));
			}
		    else
			{
			    (*it).second++;
			}
		}

	    VD.push_back(D);
//	    cout << endl;
	}

    cerr << "----------------------------------------" << endl;

//    cout << endl;
/*
    // stopwords:
    set<string>::iterator	s_it=stopwords->begin(), s_stop=stopwords->end();
    for (; s_it!=s_stop; ++s_it)
	{
	    cout << (*s_it) << " ";
	}
//    cout << endl;
*//*
    // stopwords:
    map<string,int>::iterator	it=S.begin(), stop=S.end();
    int		size = (result.size() * 4)/10;

    for (; it!=stop; ++it)
	{
	    if ((*it).second > size || (*it).second < 3)
		stopwords->insert((*it).first);
//		cout << (*it).first << " ";
	}*/
/*
//    cout << size << endl;
    for (; it!=stop;)
	{
	    old_it = it;
	    ++it;

	    if ((*old_it).second > size)
		cout << (*old_it).first << " ";
/+
	    if ((*old_it).second <= size)
		{
		    S.erase(old_it);
		}
	    else
		cout << (*old_it).first << " ";
+/
	}
    cout << endl;
*/

//    for (int doc=0; doc<result.size(); doc++)

    if (means)
	{
	    cout << "Writing dictionary: "; cout.flush();
	    string	f_words_name = means_prefix + "_words";
	    ofstream	f_words(f_words_name.c_str());

	    map< string,list<int> >::iterator	it = II.begin(), stop = II.end();
	    map<string,int>			dual;
	    vector<string>			word;
	    vector<double>			idf;
	    for (int nr=0; it!=stop; ++it,nr++)
		{
		    if (nr % 1000 == 0) { cout << "."; cout.flush(); }
		    f_words << (*it).first << endl;
		    word.push_back((*it).first);
#ifdef DUAL
		    dual.insert(pair<string,int>((*it).first,nr));
#endif
#if 0
		    string	w = (*it).first;
		    size_t	pos;
		    if ((pos=w.find(' '))>=0)
			{
			    string	w1, w2;
			    double	d1, d2;
			    w1 = w.substr(0,pos);
			    w2 = w.substr(pos+1);

			    struct idf_node	*n;
			    n = (struct idf_node*)bsearch(w1.c_str(), mmap_idf, n_idf, sizeof(struct idf_node), cmp_idf);
			    if (n==NULL) d1 = log(600000);
			    else d1 = n->f;

			    n = (struct idf_node*)bsearch(w2.c_str(), mmap_idf, n_idf, sizeof(struct idf_node), cmp_idf);
			    if (n==NULL) d2 = log(600000);
			    else d2 = n->f;

			    idf.push_back(d1+d2);
			}
		    else
#endif
			{
			    struct idf_node	*n = (struct idf_node*)bsearch((*it).first.c_str(), mmap_idf, n_idf, sizeof(struct idf_node), cmp_idf);
			    if (n==NULL)
				idf.push_back(log(600000));
			    else
			    {
//			cout << n->w << ": " << n->f << endl;
				idf.push_back(n->f);
			    }
			}
//		    double	idf = log( (double)(1+VD.size()) / (double)(*S.find((*it).first)).second );
		}

	    it = II.begin();
	    for (int nr=0; it!=stop; ++it,nr++)
		{
		    if (nr % 1000 == 0) { cout << "."; cout.flush(); }
		    f_words << (*it).first << endl;
		    f_words << nr << " " << (*it).second.size() << " " << (*it).second.size() << endl;
		}
	    f_words.close();
	    cout << endl;
	    cout << "Generating sparse matrix: "; cout.flush();

	    string	f_col_ccs_name = means_prefix + "_col_ccs",
			f_row_ccs_name = means_prefix + "_row_ccs",
			f_tfn_nz_name = means_prefix + "_tfn_nz";
	    ofstream	f_col_ccs(f_col_ccs_name.c_str()), f_row_ccs(f_row_ccs_name.c_str()), f_tfn_nz(f_tfn_nz_name.c_str());
	    int		rows, cols, nz;
	    int		rowptr=0, colptr=0;

	    rows = II.size();
	    cols = result.size();
	    nz = 0;

	    /***/
	    set<int>	*dw = new set<int>[result.size()];
	    // document, list_of_words

	    it = II.begin();
	    for (int nr=0; it!=stop; ++it,nr++)
		{
		    if (nr % 1000 == 0) { cout << "."; cout.flush(); }

		    list<int>::iterator	ii_it=(*it).second.begin(), ii_stop=(*it).second.end();
		    for (; ii_it!=ii_stop; ++ii_it)
			{
			    int		_id = *ii_it;
			    dw[_id].insert(nr);
			}
		}
	    /***/

	    f_col_ccs << "0" << endl;

	    for (int i=0; i<result.size(); i++)
		{
		    set<int>::iterator	s_it=dw[i].begin(), s_stop=dw[i].end();
		    for (; s_it!=s_stop; ++s_it)
			{

			    f_row_ccs << *s_it << endl;
//			    f_tfn_nz << "1.0" << endl;

			    if (stopwords->find(word[*s_it]) != stopwords->end())
				{
				    f_tfn_nz << "0" << endl;
				}
			    else
				{
				    string	w = word[*s_it];
				    double	tf_idf = 0;
//				    double	_idf = log( (double)(1+VD.size()) / (double)(*S.find(word[*s_it])).second );

//				    if (process == p_snippet) doc_size = VD[i]->getSnippetData()->size();
//				    else /*if (process == p_para)*/ doc_size = VD[i]->getDocData()->size();

#ifdef DUAL
				    size_t	pos;
				    if ((pos=w.find(' '))!=string::npos)
					{
					    string	w1, w2;
					    w1 = w.substr(0,pos);
					    w2 = w.substr(pos+1);

					    int		p1, p2;
					    p1 = (*dual.find(w1)).second;
					    p2 = (*dual.find(w2)).second;
/*
					    if (w == "tony blair")
						{
//						    cout << "\"" << w1 << "\", \"" << w2 << "\"" << endl;
						    cout << "tf:tony = " << (1 + log(1 + log((double)VD[i]->getWordOccurences(w1)))) << ", ";
						    cout << "idf:tony = " << idf[p1] << endl;
						    cout << "tf:blair = " << (1 + log(1 + log((double)VD[i]->getWordOccurences(w2)))) << ", ";
						    cout << "idf:blair = " << idf[p2] << endl;
						}
*/
					    if (stopwords->find(w1)==stopwords->end())
						tf_idf+= (1 + log(1 + log((double)VD[i]->getWordOccurences(w1)))) * idf[p1];

					    if (stopwords->find(w2)==stopwords->end())
						tf_idf+= (1 + log(1 + log((double)VD[i]->getWordOccurences(w2)))) * idf[p2];
					}
				    else
#endif
					{
/*
					    if (w == "tony" || w == "blair")
						{
						    cout << " tf:" << w << " = " << (1 + log(1 + log((double)VD[i]->getWordOccurences(w)))) << ", ";
						    cout << " idf:" << w << " = " << idf[*s_it] << endl;
						}
*/
					    tf_idf = (1 + log(1 + log((double)VD[i]->getWordOccurences(w)))) * idf[*s_it];
//					    tf = (double)VD[i]->getWordOccurences(word[*s_it]) / doc_size;
					}

				    double	docs_w = (*II.find(w)).second.size();

				    f_tfn_nz << tf_idf*docs_w << endl;
				    if (w == "tony" || w == "blair" || w == "tony blair")
					cout << w << ": tfidf(" << tf_idf << ") x docs(" << docs_w << ") = " << tf_idf*docs_w << endl;
				}

			    colptr++;
			}

		    f_col_ccs << colptr << endl;
		}

	    nz = colptr;

	    f_col_ccs.close();
	    f_row_ccs.close();
	    f_tfn_nz.close();

	    string	f_dim_name = means_prefix + "_dim";
	    ofstream	f_dim(f_dim_name.c_str());

	    f_dim << rows << " " << cols << " " << nz << endl;
	    f_dim.close();
	    cout << "done." << endl;
	}
    else
	{
	    for (int doc=0; doc<VD.size(); doc++)
		{
		    cout << VD[doc]->getDocid() << endl;
		    cout << VD[doc]->getUrl() << endl;
		    cout << VD[doc]->getPath() << endl;
		    cout << VD[doc]->getCategory() << endl;
		    cout << VD[doc]->getTitle() << endl;

//		    vector<string>	&V = *VD[doc]->getdataSnippet(&qa, SNIPPET_LEN);
		    vector<string>	*V;
		    if (process == p_snippet) V = VD[doc]->getSnippetData();
		    else if (process == p_para) V = VD[doc]->getDocData();

		    cout << VD[doc]->getSnippet() << endl;

		    map<string,int>	L;
		    int			total=0;
		    for (int i=0; V!=NULL && i<V->size(); i++)
			{
			    double	score;
			    string	w = (*V)[i];
			    if (total + w.size() +2 >= 1048500) break;

			    if (stopwords->find(w) != stopwords->end())
				{ score = 0; }
			    else
				{
				    double	tf, idf;
//				    tf = (double)VD[doc]->getWordOccurences(w) / (double)V->size();
//				    idf = log( (double)VD.size() / (double)(*S.find(w)).second );
//				    score = tf*idf;
				    struct idf_node	*n;
				    n = (struct idf_node*)bsearch(w.c_str(), mmap_idf, n_idf, sizeof(struct idf_node), cmp_idf);
				    if (n==NULL) idf = log(600000);
				    else idf = n->f;

				    score = idf; // Test using idf only.
				}
		    //{ cout << "_"; total++; }
			    cout << w << " " << score << " ";
			    total+= w.size() +10;
			}

		    cout << endl;
		}

    	    cout << endl;
	}
}
/*
* antall forekomster i current doc
* V.size()
* antall_doc
* VD.size()
*/

