
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <iostream>
#include <fstream>
#include <streambuf>
#include <vector>
#include <map>
#include <set>

using namespace std;



class fdinbuf : public std::streambuf {
   protected:
     int fd;    // file descriptor
   protected:
     /* data buffer:
      * - at most, pbSize characters in putback area plus
      * - at most, bufSize characters in ordinary read buffer
      */
     static const int pbSize = 4;        // size of putback area
     static const int bufSize = 1024;    // size of the data buffer
     char buffer[bufSize+pbSize];        // data buffer
 
   public:
     /* constructor
      * - initialize file descriptor
      * - initialize empty data buffer
      * - no putback area
      * => force underflow()
      */
     fdinbuf (int _fd) : fd(_fd) {
         setg (buffer+pbSize,     // beginning of putback area
               buffer+pbSize,     // read position
               buffer+pbSize);    // end position
     }
 
   protected:
     // insert new characters into the buffer
     virtual int_type underflow () {
 #ifndef _MSC_VER
         using std::memmove;
 #endif
 
         // is read position before end of buffer?
         if (gptr() < egptr()) {
             return traits_type::to_int_type(*gptr());
         }
 
         /* process size of putback area
          * - use number of characters read
          * - but at most size of putback area
          */
         int numPutback;
         numPutback = gptr() - eback();
         if (numPutback > pbSize) {
             numPutback = pbSize;
         }
 
         /* copy up to pbSize characters previously read into
          * the putback area
          */
         memmove (buffer+(pbSize-numPutback), gptr()-numPutback,
                 numPutback);
 
         // read at most bufSize new characters
         int num;
         num = read (fd, buffer+pbSize, bufSize);
         if (num <= 0) {
             // ERROR or EOF
             return EOF;
         }
 
         // reset buffer pointers
         setg (buffer+(pbSize-numPutback),   // beginning of putback area
               buffer+pbSize,                // read position
               buffer+pbSize+num);           // end of buffer
 
         // return next character
         return traits_type::to_int_type(*gptr());
     }
 };
 
 class fdistream : public std::istream {
   protected:
     fdinbuf buf;
   public:
     fdistream (int fd) : std::istream(0), buf(fd) {
         rdbuf(&buf);
     }
 };




class fdoutbuf : public std::streambuf
{
    public:

	fdoutbuf( int _fd )
	: fd(_fd)
	    {
	    }

    protected:

	int	fd;

	virtual int_type overflow( int_type c )
	    {
		if (c!=EOF)
		    {
			char z = c;
			if (write(fd, &z, 1) != 1)
			    return EOF;
		    }

		return c;
	    }

	std::streamsize xsputn( const char *s, std::streamsize num )
	    {
		return write(fd,s,num);
	    }
};


class fdostream : public std::ostream
{
    public:

	fdostream( int fd )
	: std::ostream(0), buf(fd)
	    {
		rdbuf(&buf);
	    }

    protected:

	fdoutbuf	buf;
};



class Server
{
    public:

	Server( int _maxclients )
	: maxclients(_maxclients)
	    {
		int		opt=1;

		clientsd = new int[maxclients];
		for (int i=0; i<maxclients; i++)
		    clientsd[i] = 0;

		num_clients = 0;

	        sd=socket(AF_INET,SOCK_STREAM,0);
	        if (sd==-1)
		    {
		        perror("socket");
			exit(EXIT_FAILURE);
		    }

		if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (char *)&opt, sizeof(opt))<0)
		    {
			perror("setsockopt");
			exit(EXIT_FAILURE);
		    }

		address.sin_family = AF_INET;
	        address.sin_addr.s_addr = INADDR_ANY;
	        address.sin_port = htons(7000);
	        addrlen = sizeof(address);
	        if (bind(sd, (struct sockaddr *)&address, addrlen)<0)
	    	    {
	    		perror("bind");
	    	        exit(EXIT_FAILURE);
	    	    }

	        if (listen(sd, maxclients)<0)
	    	    {
			perror("listen");
			exit(EXIT_FAILURE);
		    }

		cout << "Waiting for connections. Enter 'start' to begin." << endl;

	        for (;;)
	    	    {
	    		int		activity;

			FD_ZERO(&readfds);

			FD_SET(sd, &readfds);
			FD_SET(STDIN_FILENO, &readfds);

//			struct timeval	timeout;
			activity = select(maxclients+3, &readfds, NULL, NULL, NULL);

//			if ((activity < 0) && (errno!=EINTR))
//			    {
//			    }

			if (FD_ISSET(sd, &readfds))
			    {
				int		newsd;

			        if ((newsd = accept(sd, (struct sockaddr *)&address, (socklen_t*)&addrlen))<0)
				    {
					perror("accept");
				        exit(EXIT_FAILURE);
				    }

				clientsd[num_clients++] = newsd;
				cout << "New connection(" << num_clients-1 << ")" << endl;
				client_in.push_back(new fdistream(newsd));
				client_out.push_back(new fdostream(newsd));
			    }

			if (FD_ISSET(STDIN_FILENO, &readfds))
			    {
				string		command;
				cin >> command;
				if (command=="q" || command=="quit")
				    {
					for (int i=0; i<maxclients; i++)
					    if (clientsd[i]>0)
						close(clientsd[i]);

					close(sd);
					exit(EXIT_SUCCESS);
				    }
				else if (command=="s" || command=="start")
				    {
					cout << "Starting with " << num_clients << " connections." << endl;
					return;
				    }
				else cout << "Unknown command: " << command << endl;
			    }
		    }
	    }


	~Server()
	    {
		for (int i=0; i<maxclients; i++)
		    if (clientsd[i]>0)
			{
			    delete client_in[i];
			    delete client_out[i];
			    close(clientsd[i]);
			}

		close(sd);
	    }

	int numConnections()
	    {
		return num_clients;
	    }

	void addDocument( int client, int docid, vector<string> V, vector<double> D )
	    {
//		if (docid==0) cout << "addDocument: size=" << V.size() << endl;
//		if (docid==0) cout << "addDocument:";

		(*client_out[client]) << docid << " " << V.size() << endl;
		for (int i=0; i<V.size(); i++)
		    {
			(*client_out[client]) << V[i] << " " << D[i] << " ";
//			if (docid==0) cout << " " << V[i];
		    }
		(*client_out[client]) << endl;
//		if (docid==0) cout << endl;
	    }

	void sendFinalBC( int client, vector< vector<string> > FinalBC)
	    {
		(*client_out[client]) << FinalBC.size() << endl;
		for (int i=0; i<FinalBC.size(); i++)
		    {
			(*client_out[client]) << FinalBC[i].size();
			for (int j=0; j<FinalBC[i].size(); j++)
			    (*client_out[client]) << " " << FinalBC[i][j];
			(*client_out[client]) << endl;
		    }
	    }

	void sendInt( int client, int value )
	    {
		(*client_out[client]) << value << endl;
	    }

	void sendThreshold( int client, double t )
	    {
		(*client_out[client]) << t << endl;
	    }

	vector< set<int> > getDoclist( int client )
	    {
		vector< set<int> >	Vi;
		int		size;

		(*client_in[client]) >> size;

		for (int i=0; i<size; i++)
		    {
			int	p_size;
			int	val;
			set<int>	vi;

			(*client_in[client]) >> p_size;

			for (int j=0; j<p_size; j++)
			    {
				(*client_in[client]) >> val;
				vi.insert(val);
			    }

			Vi.push_back(vi);
		    }

		return Vi;
	    }

	map<vector<string>, pair<int,double> > getBCPhrases( int client )
	    {
		int		size;
		map<vector<string>, pair<int,double> >	M;

		(*client_in[client]) >> size;
		for (int i=0; i<size; i++)
		    {
			double	d;
			int	sz, p_size;
			string	s;
			vector<string>	vs;

			(*client_in[client]) >> sz >> d >> p_size;
			for (int j=0; j<p_size; j++)
			    {
				(*client_in[client]) >> s;
				vs.push_back(s);
			    }

			M.insert( pair<vector<string>, pair<int,double> >(vs, pair<int,double>(sz,d)) );
		    }

		return M;
	    }

	void chat()
	    {
		for (;;)
		    {
	    		int		activity;

			FD_ZERO(&readfds);

			FD_SET(STDIN_FILENO, &readfds);
			for (int i=0; i<num_clients; i++)
			    {
				if (clientsd[i] > 0)
				    {
					FD_SET(clientsd[i], &readfds);
    				    }
			    }

//			struct timeval	timeout;
			activity = select(num_clients+4, &readfds, NULL, NULL, NULL);

//			if ((activity < 0) && (errno!=EINTR))
//			    {
//			    }

		        int		size;
		        char		rcv_buffer[8192], snd_buffer[8192];

		        for (int i=0; i<num_clients; i++)
			    if (FD_ISSET(clientsd[i], &readfds))
//				if ((size = read(clientsd[i], rcv_buffer, 8191)) < 0)
				if ((*client_in[i]) >> rcv_buffer)
				    {
					rcv_buffer[size-1] = '\0';
					snprintf(snd_buffer, 8192, "  %i: %s", i, rcv_buffer);

				        for (int j=0; j<num_clients; j++)
					    if (i!=j)
						{
						    (*client_out[j]) << snd_buffer << endl;
					    	    //send(clientsd[j], snd_buffer, strlen(snd_buffer), 0);
				    		}
				    }
				else
				    {
					cerr << "Transmission error. Closing client " << i << endl;
				        close(clientsd[i]);
				        clientsd[i] = 0;
				    }

			if (FD_ISSET(STDIN_FILENO, &readfds))
			    {
				string		command;
				cin >> command;
				if (command=="q" || command=="quit")
				    {
					return;
				    }
				else cout << "Unknown command: " << command << endl;
			    }
		    }
	    }

    private:

	int			sd, addrlen, maxclients, num_clients;
	struct sockaddr_in	address;
	int			*clientsd;
	fd_set			readfds;
	vector<fdistream*>	client_in;
	vector<fdostream*>	client_out;
};


class Client
{
    public:

	Client( char *hostname )
	    {
		struct hostent	*he;

		if ((he=gethostbyname(hostname)) == NULL)
		    {
			perror("gethostbyname");
			exit(EXIT_FAILURE);
		    }

		if ((sd=socket(PF_INET, SOCK_STREAM, 0)) == -1)
		    {
			perror("socket");
			exit(EXIT_FAILURE);
		    }

		address.sin_family = AF_INET;    // host byte order 
		address.sin_port = htons(7000);  // short, network byte order 
	        address.sin_addr = *((struct in_addr *)he->h_addr);
	        memset(address.sin_zero, '\0', sizeof(address.sin_zero));

		if (connect(sd, (struct sockaddr *)&address, sizeof(address)) == -1)
		    {
			perror("connect");
			exit(EXIT_FAILURE);
		    }

		cout << "Connected to " << hostname << endl;

		client_in = new fdistream(sd);
		client_out = new fdostream(sd);
	    }


	~Client()
	    {
		delete client_in;
		delete client_out;
		close(sd);
	    }

	double getThreshold()
	    {
		double	t;
		(*client_in) >> t;
		return t;
	    }

	bool getDocument( int *docid, vector<string> *_V, vector<double> *_D )
	    {
		int		size;
		string		s;
		double		v;
		vector<string>	&V = *_V;
		vector<double>	&D = *_D;

//		cout << "Recieving data"; cout.flush();
		(*client_in) >> *docid >> size;
//		cout << "."; cout.flush();

//		if (*docid == 0) cout << "getDocument: size=" << size << endl;

		if (size < 0) return false;

		for (int i=0; i<size; i++)
		    {
			(*client_in) >> s >> v;
			V.push_back(s);
			D.push_back(v);
//			cout << "."; cout.flush();
		    }
//		cout << endl;

		return true;
	    }

	void getFinalBC( vector< vector<string> > *FinalBC )
	    {
		int		size;
		(*client_in) >> size;

		for (int i=0; i<size; i++)
		    {
			int		p_size;
			string		s;
			vector<string>	V;

			(*client_in) >> p_size;

			for (int j=0; j<p_size; j++)
			    {
				(*client_in) >> s;
				V.push_back(s);
			    }

			FinalBC->push_back(V);
		    }
	    }

	void sendDoclist( vector< set<int> > Vi )
	    {
		(*client_out) << Vi.size() << endl;

		for (int i=0; i<Vi.size(); i++)
		    {
			(*client_out) << Vi[i].size();
			set<int>::iterator	it=Vi[i].begin(), stop=Vi[i].end();
			for (; it!=stop; ++it)
			    (*client_out) << " " << *it;
			(*client_out) << endl;
		    }
	    }

	void sendBCPhrases(vector< pair< pair<int,double>, vector<string> > > V)
	    {
		(*client_out) << V.size() << endl;

		for (int i=0; i<V.size(); i++)
		    {
			(*client_out) << V[i].first.first << " " << V[i].first.second << " " << V[i].second.size() << endl;
			for (int j=0; j<V[i].second.size(); j++)
			    {
				(*client_out) << V[i].second[j] << " ";
			    }
			(*client_out) << endl;
		    }
	    }

    private:

	int			sd;
	struct sockaddr_in	address;
	fdistream*		client_in;
	fdostream*		client_out;
};


/*

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>

#define PORT 3490 // the port client will be connecting to 

#define MAXDATASIZE 100 // max number of bytes we can get at once 

int main(int argc, char *argv[])
{
    int sockfd, numbytes;  
    char buf[MAXDATASIZE];
    struct hostent *he;
    struct sockaddr_in their_addr; // connector's address information 

    if (argc != 2) {
        fprintf(stderr,"usage: client hostname\n");
        exit(1);
    }

    if ((he=gethostbyname(argv[1])) == NULL) {  // get the host info 
        herror("gethostbyname");
        exit(1);
    }

    if ((sockfd = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
        perror("socket");
        exit(1);
    }

    their_addr.sin_family = AF_INET;    // host byte order 
    their_addr.sin_port = htons(PORT);  // short, network byte order 
    their_addr.sin_addr = *((struct in_addr *)he->h_addr);
    memset(their_addr.sin_zero, '\0', sizeof their_addr.sin_zero);

    if (connect(sockfd, (struct sockaddr *)&their_addr,
                                          sizeof their_addr) == -1) {
        perror("connect");
        exit(1);
    }

    if ((numbytes=recv(sockfd, buf, MAXDATASIZE-1, 0)) == -1) {
        perror("recv");
        exit(1);
    }

    buf[numbytes] = '\0';

    printf("Received: %s",buf);

    close(sockfd);

    return 0;
} 
*/

/*
int main(int argc, char *argv[])
{
#ifdef SERVER
    Server	*server = new Server(16);

    server->chat();

    delete server;
#else
    Client	*client = new Client("localhost");
#endif
}
*/

