
/**
 *	Simple search-system, (C) Copyright Magnus Galåen
 */

#define SIMPLE

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <list>

extern "C"
{
#include "query/query_parser.h"
}

#include "blocks.h"
#include "output.h"

using namespace std;



int main( int argc, const char *argv[] )
{
#ifdef SIMPLE
    if (argc<3)
	{
	    cout << "Usage: " << argv[0] << " <query> <fasit.X>" << endl;
	    exit(-1);
	}
#else
    if (argc<5)
	{
	    cout << "Usage: " << argv[0] << " <matrix_words> <matrix_tfn_doctoclus.X> <matrix_tfn_wordtoclus.X> <query>" << endl;
	    exit(-1);
	}
#endif

    FILE	*f_id = fopen("id", "r");
    int		size_id;
    void	*mmap_id;
    struct stat	inode;

    fstat(fileno(f_id), &inode);
    size_id = inode.st_size;

    mmap_id = mmap(0, size_id, PROT_READ, MAP_PRIVATE, fileno(f_id), 0);

    const char	*ptr = &argv[2][strlen(argv[2])-1];
    while (*ptr!='.') ptr--;
    ptr++;

    int		num_clusters = atoi(ptr), num_documents = 0;
    cout << num_clusters << " clusters." << endl;
    list<int>	C[num_clusters];
    ifstream	f_clusters(argv[2]);
#ifndef SIMPLE
    ifstrem	f_words(argv[1]), f_wordtoclus(argv[3]);
#endif
    int		cnr, docid;
/*
    for (int c=0; c<num_clusters; c++)
	{
	    C.push_back(new list<int>());
	}
*/

#ifndef SIMPLE
    f_clusters >> docid;
#endif

    while (f_clusters >> cnr >> docid)
	{
	    C[cnr].push_back(docid);
	    num_documents++;
	}

#ifndef SIMPLE
    int		wsize;
//    vector<string>	dict;
    set<string>		W[num_clusters];
    f_wordtoclus >> wsize;
    for (int i=0; i<wsize; i++)
	{
	    char	buf[256];
	    string	s;
	    int		x;
	    f_words.getline(buf, 255);
	    s = buf;
//	    dict.push_back(s);

	    f_wordtoclus >> x;
	    W[x].insert(s);
	}

    query_array		qa;
    get_query((char*)argv[4], strlen((char*)argv[4]), &qa);
#else
    query_array		qa;
    get_query((char*)argv[1], strlen((char*)argv[1]), &qa);
#endif

    Output	*output = new Output("answer.html", &qa, mmap_id);
    output->print_sidebar(num_documents);

    for (int c=0; c<num_clusters; c++)
	{
#ifdef SIMPLE
	    output->print_snippets(C[c]);
#else
	    output->print_snippets(C[c], W[c]);
#endif
	}

    delete output;
}

