
#include <assert.h>
#include <limits.h>
#include <math.h>
#include <sys/timeb.h>

#include <iostream>
#include <iomanip>
#include <fstream>
#include <list>
#include <vector>
#include <set>
#include <map>
#include <string>
#include <sstream>

#include "document.cpp"

//#define THRESHOLD

#if defined(SERVER) || defined(CLIENT)
#include "server.cpp"
#endif

extern "C"
{
#include "query/query_parser.h"
#include "generateSnippet/snippet.parser.h"
}

using namespace std;



struct doc
{
    int		docid;
    char	*url, *path, *category, *title, *snippet;
    vector<string>	words;
    vector<double>	tfidf;
};

vector<struct doc>	Doc;


typedef struct
{
    int		k, p;
    double	score;
//    vector<bool>	stopw;
} basecluster;

typedef struct
{
    set< pair<int,int> >	docs;
//    vector< pair<int,int> >	P;
    multimap<int, basecluster>	P;
} single_cluster;


typedef struct
{
    int				docs_in_r;
    vector<single_cluster>	R;
} cluster_result;



string print_strings( vector<string> V )
{
    ostringstream	oss;

    int i;
    for (i=0; i<V.size() && i<10; i++)
	{
	    if (i>0) oss << " ";
	    oss << V[i];
	}

    if (i<V.size()) oss << " ...";

    return oss.str();
}

char	*addr_name[] = {"alpha","beta","gamma","delta","epsilon","zeta","eta","theta","iota","lambda","mu","nu","xi","omicron","phi","rho","sigma","tau","upsilon","omega",
			"anker","gul","kaffe","snelle","kantarell","tralle","techno"};
map<int,char*>	addr_hash;
int	last=0;

string print_addr( void *ptr )
{
//    return string("<addr>");
    map<int,char*>::iterator	it=addr_hash.find((int)ptr);

    if (it==addr_hash.end())
	{
	    addr_hash.insert(pair<int,char*>((int)ptr, addr_name[last]));
	    return string(addr_name[last++]);
	}
    else
	{
	    return string((*it).second);
	}
}


class SuffixTree;

class SuffixTree
{
    public:

	class Node;
	vector<string>	T;
	vector<double>	tfidf;
	int		doc_nr;
        typedef struct
	    {
		int	k, p;
		Node*	s;
	    } transition;

	class Node
	    {
		public:

		    Node(SuffixTree *_parent)
			: parent(_parent)
			{
			    slink = NULL;

//			    if (doc_nr>=0)
//				docs.insert(doc_nr);
			}

		    ~Node()
			{
			}

		    int compare(Node *s)
			{
			    set< pair<int,int> >::iterator	this_it=docs.begin(), this_stop=docs.end(),
						s_it=s->docs.begin(), s_stop=s->docs.end();

			    int		count = 0;

			    for (; this_it!=this_stop && s_it!=s_stop; ++this_it)
				{
				    while (s_it!=s_stop && (*s_it).first < (*this_it).first) ++s_it;

				    if ((*s_it).first == (*this_it).first) count++;
				}

//			    if (count > docs.size()/2) return true;	// ---- (fikset): docs.size() kan inneholde duplikate dokumenter :p
//			    return false;
//			    return (count*100) / docs.size();
			    return count;
			}

		    void create_transition(int k, int p, Node *s)
			{
			    transition	t;
			    t.k = k;
			    t.p = p;
			    t.s = s;

			    d[parent->T[k]] = t;
//			    t.s->add_doc_at(k);
			}

		    void add_doc(int doc_nr, int k)
//		    void add_doc(int doc_nr)
			{
//			    cout << doc_nr << "," << k << ":";
//			    for (int i=0; i<5 && k+i<parent->T.size(); i++) cout << " " << parent->T[k+i];
//			    cout << endl;
			    docs.insert( pair<int,int>(doc_nr, k) );
//			    docs.insert(doc_nr);
			}

		    Node* find(vector<string> V)
			{
			    map<string, transition>::iterator	it=d.find(V[0]);

			    if (it==d.end()) return NULL;

			    transition	t1 = (*it).second;
			    if (V.size() < t1.p-t1.k+1) return NULL;

			    for (int i=t1.k; i<=t1.p; i++)
				{
				    if (V[i-t1.k] != parent->T[i]) return NULL;
				}

			    if (V.size() == t1.p-t1.k+1) return t1.s;

			    vector<string>	NV;
			    for (int i=t1.p-t1.k+1; i<V.size(); i++)
				NV.push_back(V[i]);

			    return t1.s->find(NV);
			}

		    void print_all(int indent=0)
			{
			    map< string, transition >::iterator	it=d.begin(), stop=d.end();
			    for (; it!=stop; ++it)
				{
				    string		t = (*it).first;
				    transition		t1 = (*it).second;

				    for (int i=0; i<indent; i++) cout << " ";
				    cout << "|";

//				    cout << " (" << t << ")(" << t1.k << "," << t1.p << ")";
				    for (int i=t1.k; i<=t1.p && i<parent->T.size(); i++) cout << " " << parent->T[i];

				    set< pair<int,int> >::iterator	s_it=t1.s->docs.begin(), s_stop=t1.s->docs.end();
				    for (; s_it!=s_stop; ++s_it) cout << " " << (*s_it).first;// << "," << (*s_it).second;
//				    cout << " (" << print_addr(s) << ")";

				    cout << endl;
				    t1.s->print_all(indent + t1.p - t1.k + 2);
				}
			}

		    void erase()
			{
			    map< string, transition >::iterator	it=d.begin(), stop=d.end();
			    for (; it!=stop; ++it)
				{
				    transition		t1 = (*it).second;
				    t1.s->erase();
				    delete t1.s;
				}
			}

		    void strip()
			{
			    map< string, transition >::iterator	it=d.begin(), stop=d.end();
			    for (; it!=stop;)
				{
				    string		t = (*it).first;
				    transition		t1 = (*it).second;

				    if (t=="$")
					{
					    t1.s->erase();
					    delete t1.s;
					    map<string,transition>::iterator old_it = it;
					    ++it;
					    d.erase(old_it);
					    continue;
					}

				    bool	delete_children = false;

				    for (int i=t1.k; i<=t1.p && i<parent->T.size(); i++)
					if (parent->T[i] == "$")
					    {
						(*it).second.p = i-1;
						delete_children = true;
						break;
					    }

				    if (delete_children)
					{
					    t1.s->erase();
					    t1.s->d.erase(t1.s->d.begin(), t1.s->d.end());
					}
				    else
					{
					    t1.s->strip();
					}

				    ++it;
				}
			}

		    void traverse( int k, int p )
			{
//			    add_doc(parent->get_doc_at(k),k);
/*
			    cout << "[" << docs.size() << "]";
			    for (int i=k; i<p; i++)
				cout << " " << parent->T[i];
			    cout << endl;
*/
//			    if (docs.size() > 3)// && docs.size() <= (parent->doc_nr*0.4))
//			    {


#define IDF_SCORE
#ifdef IDF_SCORE
			    double		score;
			    double		bonus = 0.0;
			    double		termfreq = 0.0;
			    int			size = 0;

			    for (int i=k; i<=p; i++)
				{
				    if (parent->tfidf[i]!=0)
					{
					    size++;
					    termfreq+= parent->tfidf[i];
					}
				    else bonus+= 0.001;
				}
/*
			    double		penalty = 0.4 - ((log((double)parent->doc_nr) -3) *4)/10;

			    if (penalty < 0.05) penalty = 0.05;
			    else if (penalty > 0.4) penalty = 0.4;
*/
			    if (size==0) score = 0.0;
			    else if (size==1) score = 0.3 + (termfreq/10);
//			    else if (size<=2) score = termfreq/1.5;
//			    else if (size<=3) score = termfreq/2.0;
//			    else if (size<=4) score = termfreq/2.5;
//			    else if (size<=5) score = termfreq/3.0;
//			    else if (size<=6) score = termfreq/3.5;
//			    else score = termfreq / (double)size;
			    else if (size<=6) score = 1.75 + ((double)size/2) + (termfreq/((double)size*10));
			    else score = 4.75 + ((termfreq*6)/((double)size*10));
//			    else score = 0.0;

			    score+= bonus;
    	    		    score*= (double)docs.size();
#else
			    double		score = 1.0;
// 3 doc x 2 len => 8.25
// 4 doc x 2 len => 11
// 2 doc x 6 len => 9.5
// 3 doc x 6 len => 14.25
			    int		size = 0;
			    double	bonus = 0.0;
//			    double	termfreq = 0.0;
			    for (int i=k; i<=p; i++)
				{
//				    char	buf[1024];
//				    strcpy(buf, parent->T[i].c_str());
//				    convert_to_lowercase((unsigned char*)buf);
//				    set<string>::iterator	it=S.find(string(buf));
//				    if (it==S.end()) size++;
//				    if (parent->T[i][0]!='_') size++;
				    if (parent->tfidf[i]!=0)
					{
					    size++;
//					    termfreq+= parent->tfidf[i];
					}
				    else bonus+= 0.001;
				}

			    if (size==0) score = 0.0;
			    else if (size==1) score = 0.30;	// Bør penalty variere med størrelsen på dokumentsamlingen?
			    else if (size<=6) score = 1.75 + ((double)size)/2;
			    else score = 4.75;

			    score+= bonus;
    	    		    score*= ((double)docs.size());
//			    score*= (double)size*termfreq;	// For dårlige resultat!!!
#endif
//			    if (0==1 && score >= 20)
//			    {
//			    cout << "{" << score << "} <= ";
//			    for (int i=0; i<P.size(); i++)
//				    cout << P[i] << " ";
//			    for (int i=k; i<=p; i++)
//				cout << parent->T[i] << " ";
//			    cout << "[" << docs.size() << "," << size << "]" << endl;
//			    }

			    transition	t2;
			    t2.k = k;
			    t2.p = p;
			    t2.s = this;
			    parent->B.insert(pair<double, transition>(score, t2));

//			    }
/*
			    cout << "(" << docs.size() << ")";
			    for (int i=0; i<P.size(); i++)
				{
				    cout << " " << P[i];
//				    if (i>0) cout << ",";
//				    for (int j=P[i].first; j<P[i].second; j++)
//					cout << " " << parent->T[j];
				}
			    cout << endl;
//			    cout << endl;
*/
			    map< string, transition >::iterator	it=d.begin(), stop=d.end();
			    for (; it!=stop; ++it)
				{
				    transition		t1 = (*it).second;
//				    vector<string>	O;
//				    O.insert(O.end(), P.begin(), P.end());
//				    string		s((*it).first);
//				    s = "[" + s + "]";
//				    O.push_back(s);
//				    for (int i=t1.k; i<=t1.p; i++)
//					O.push_back( parent->T[i] );
				    t1.s->traverse(t1.k - (p - k + 1), t1.p );
				}
			}

/*
		    list< pair< vector<string>, set<int> > > traverse()
			{
			    vector<string>	empty;
			    return traverse(empty);
			}


		    list< pair< vector<string>, set<int> > > traverse(vector<string> prefix)
			{
			    list< pair< vector<string>, set<int> > >	E;
			    map< string, pair<vector<string>, Node*> >::iterator	it=d.begin(), stop=d.end();
			    for (; it!=stop; ++it)
				{
				    vector<string>	T1 = (*it).second.first;
				    Node		*s = (*it).second.second;

				    if (s->docs.size() >= 2)
					{
					    vector<string>	newT1;
					    for (int i=0; i<prefix.size(); i++) newT1.push_back(prefix[i]);
					    for (int i=0; i<T1.size(); i++) newT1.push_back(T1[i]);

					    E.push_back( pair< vector<string>, set<int> >(newT1, s->docs) );
					    // E = E + s.traverse(newT1);
					    list< pair< vector<string>, set<int> > >	X;
					    X = s->traverse(newT1);
					    E.splice( E.end(), X );
					}
				}

			    return E;
			}

		    void print_one()
			{
			    cout << "(  ";
			    map< string, pair<vector<string>, Node*> >::iterator	it=d.begin(), stop=d.end();
			    for (; it!=stop; ++it)
				{
				    string	s = (*it).first;
				    cout << s << " ";
				}
			    cout << ")";
			}
*/
//		    map< string, pair<vector<string>, Node*> >	d;

		    SuffixTree	*parent;
		    map<string, transition>	d;
		    Node	*slink;
		    set< pair<int,int> >	docs;
//		    set<int>	docs;
	    };


	SuffixTree()
	    {
		doc_nr = 0;
		root = new Node(this);
		bottom = new Node(this);

		root->slink = bottom;

//		print_addr(root);
//		print_addr(bottom);
	    }

	~SuffixTree()
	    {
		// TODO.
	    }


	void addDocument(int docid, vector<string> _T, vector<double> _tfidf)
	    {
		real_docid.push_back(docid);
//		if (_T.size() == 0) return;

//		cout << doc_nr << "," << T.size() << ": " << print_strings(_T) << endl;

		T.insert(T.end(), _T.begin(), _T.end());
		T.push_back("$");
		tfidf.insert(tfidf.end(), _tfidf.begin(), _tfidf.end());
		tfidf.push_back(0);
		pos_doc.push_back(T.size()-1);

//		cout << "D:" << doc_nr << " " << print_strings(_T) << endl;
		doc_nr++;
	    }


	void makeGST()
	    {
		algorithm2();
		root->strip();
/*
		for (int i=0; i<T.size() && i<1500; i++)
		    {
			cout << T[i] << "," << get_doc_at(i) << "; ";
			if (T[i]=="$") cout << endl;
		    }
		cout << endl;
*/
		for (int i=0; i<T.size(); i++)
		    {
			Node	*s = root;

			map<string, transition>::iterator	it;
/*
			if (get_doc_at(i) == 1338)
			    {
				cout << "1338:";
				for (int j=0; j<10; j++)
				    cout << " " << T[i+j];
				cout << endl;
			    }
*/
			for (int j=0; T[i+j]!="$";)
			    {
				it = s->d.find(T[i+j]);
				transition	t1 = it->second;
				j+= t1.p - t1.k +1;
				s = t1.s;

				int	doc = get_doc_at(i);
				s->add_doc(doc,doc);
//				s->add_doc(get_doc_at(i),i);
			    }
		    }
	    }



	void print()
	    {
		root->print_all();
	    }


	vector< set<int> > extract(vector< vector<string> > BC)
	    {
		vector< set<int> >	Doclist;

		for (int i=0; i<BC.size(); i++)
		    {
			Node*	n = root->find(BC[i]);
			set<int>	Vi;

			if (n!=NULL)
			    {
/*
				double	score = 1.0;
			        int		size = 0;
			        double	bonus = 0.0;

			        for (int j=0; j<BC[i].size(); j++)
				    {
					if (BC[i][j][0]!='_') size++;
					else bonus+= 0.001;
				    }

				if (size==0) score = 0.0;
			        else if (size==1) score = 0.30;
			        else if (size<=6) score = 1.75 + ((double)size)/2;
			        else score = 4.75;

				score+= bonus;

    	    		        score*= ((double)n->docs.size());
				cout << score << ",";
*/
/*
				cout << n->docs.size() << ": ";

				cout << "\"" << BC[i][0];
				for (int j=1; j<BC[i].size(); j++)
				    cout << " " << BC[i][j];
				cout << "\"" << endl;
*/
				set< pair<int,int> >::iterator	it = n->docs.begin(), stop = n->docs.end();
				for (; it!=stop; ++it)
				    {
					Vi.insert((*it).first);
				    }
			    }

			Doclist.push_back(Vi);
		    }

		return Doclist;
	    }

	cluster_result baseClusters()
	    {
		cluster_result		Result;

		B.clear();
		cout << "Calculating base clusters..."; cout.flush();
		root->traverse(0, -1);
		cout << "done" << endl;

		vector<transition>	C;
		vector<double>		Score;
		vector<int>		belonging;
		set<int>		clusters;
		multimap<double, transition>::iterator	it=B.end(), stop=B.begin();
		int			b;

		cout << B.size() << " baseclusters." << endl;

//#define MANUELL_1
#ifdef MANUELL_1
		ofstream	of("ground_truth");

		for (b=0; it!=stop && b<1000; b++)
		    {
			--it;
			of << -1 << " ";
			for (int j=(*it).second.k; j<=(*it).second.p; j++)
			    of << " " << T[j];
			of << " |";
			set< pair<int,int> >::iterator d_it=(*it).second.s->docs.begin(), d_stop=(*it).second.s->docs.end();
			for (; d_it!=d_stop; ++d_it)
			    {
				of << " " << Doc[(*d_it).first].docid;
			    }
			of << endl;
		    }

		exit(0);
#endif

#if defined(CLIENT) || defined(SERVER)
		Result.docs_in_r = 0;
#endif

#if !defined(THRESHOLD)
		for (b=0; it!=stop && b<500; b++)
#else
		for (b=0; it!=stop; b++)
#endif
		    {
			--it;

//			cout << (*it).first << ":";
//			for (int j=(*it).second.k; j<=(*it).second.p; j++)
//			    cout << " " << T[j];
//			cout << endl;

			if ((*it).second.s->docs.size() < 3 || (*it).second.s->docs.size() > (double)doc_nr*0.4) continue;
#ifdef THRESHOLD
			if ((*it).first < threshold) continue;
#endif
#if defined(CLIENT) || defined(SERVER)
			single_cluster	SC;
			basecluster	bc;
			bc.k = (*it).second.k;
			bc.p = (*it).second.p;
			bc.score = (*it).first;
			SC.P.insert( pair<int, basecluster>((*it).second.s->docs.size(), bc) );
			SC.docs.insert((*it).second.s->docs.begin(), (*it).second.s->docs.end());

			Result.docs_in_r+= SC.docs.size();
			Result.R.push_back(SC);
#else
//			if ((*it).second.s->docs.size() <= 3) break;
			C.push_back((*it).second);
			Score.push_back((*it).first);
			belonging.push_back(b);
#endif
//			cout << (*it).first << ":";
//			for (int j=(*it).second.k; j<=(*it).second.p; j++)
//			    cout << " " << T[j];
//			cout << endl;

//			clusters.insert(b);
/*
			transition	t1 = (*it).second;
			cout << (*it).first << ":";
			for (int i=t1.k; i<=t1.p; i++)
			    cout << " " << T[i];
			cout << endl;
*/
		    }
#if !defined(CLIENT) && !defined(SERVER)
		cout << C.size() << " top baseclusters have been chosen." << endl;
		cout << "Calculating clusters..." << endl;
		for (int i=0; i<C.size(); i++)
		    {
			for (int j=i+1; j<C.size(); j++)
			    {
				int		count = C[i].s->compare(C[j].s);
//				if ((((count*100)/C[i].s->docs.size())>=90 ||
//				    ((count*100)/C[j].s->docs.size())>=90)
//				 || (((count*100)/C[i].s->docs.size())>50 &&
//				    ((count*100)/C[j].s->docs.size())>50))
				if (((count*100)/C[i].s->docs.size())>50 &&
				    ((count*100)/C[j].s->docs.size())>50)
				    {
					belonging[j] = belonging[i];
					clusters.insert(belonging[i]);
				    }
			    }
		    }

		cout << "Saving clusters..." << endl;
		if (clusters.size() > 0)
		    {
			Result.docs_in_r = 0;

			set<int>::iterator	it=clusters.begin(), stop=clusters.end();
			for (int x=0; it!=stop; ++it,x++)
			    {
				single_cluster	SC;

//				cout << "Cluster #" << *it << ":" << endl;
				for (int i=0; i<belonging.size(); i++)
				    {
					if (belonging[i] == *it)
					    {
//						for (int j=C[i].k; j<=C[i].p; j++)
//						    cout << T[j] << " ";
//						cout << endl;
//						SC.P.push_back( pair<int,int>(C[i].k, C[i].p) );
						basecluster	bc;
						bc.k = C[i].k;
						bc.p = C[i].p;
						bc.score = Score[i];
						SC.P.insert( pair<int, basecluster>(C[i].s->docs.size(), bc) );
						SC.docs.insert(C[i].s->docs.begin(), C[i].s->docs.end());
					    }
				    }

//				cout << "Docs:";
//				set< pair<int,int> >::iterator	d_it=SC.docs.begin(), d_stop=SC.docs.end();
//				for (; d_it!=d_stop; ++d_it)
//				    {
//					cout << " " << (*d_it).first;// << "," << (*d_it).second;
//				    }
//				cout << endl;
				Result.docs_in_r+= SC.docs.size();
				Result.R.push_back(SC);

//				cout << endl;
			    }
		    }
		else
		    {
			cout << "No clusters!" << endl;
		    }
#endif

		return Result;
/*
			it_j = B.end();
			for (int b_j=0; it_j!=stop && b_j<50; b_j++)
			    {
				--it_j;
				transition	t_j = (*it_j).second;

				if (t_i.s->compare(t_j.s) && t_j.s->compare(t_i.s))
				    {
					for (int i=t_i.k; i<=t_i.p; i++)
					    cout << T[i] << " ";
					cout << endl;
					for (int i=t_j.k; i<=t_j.p; i++)
					    cout << T[i] << " ";
					cout << endl << endl;
				    }
			    }
		    }
*/
	    }

/*
	list< pair< vector<string>, set<int> > > calculate_base_clusters()
	    {
		return root->traverse();
	    }
*/

	void setThreshold( double _t )
	    {
		threshold = _t;
	    }


    private:	/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


	void algorithm2()
	    {
		cout << "algorithm2: " << T.size() << " words" << endl;

		for (int j=0; j<T.size(); j++)
		    {
			// create transition g'(bottom, (-j,-j)) = root
			bottom->create_transition(j, j, root);
		    }
		// create suffix link f'(root) = bottom

		Node	*s = root;
		int	k = 0;

		for (int i=0; i<T.size(); i++)
		    {
			pair<Node*,int>	ret;
			ret = update(s, k, i);
//			s->add_doc(get_doc_at(k));
//			cout << "1";
//			for (int j=k; j<=i; j++)
//			    cout << T[j] << " ";
//			cout << endl;
			s = ret.first;
			k = ret.second;
//			cout << "|";
//			for (int j=k; j<=i; j++)
//			    cout << T[j] << " ";
//			cout << endl;
//			s->add_doc(get_doc_at(k));

			ret = canonize(s, k, i);
			s = ret.first;
			k = ret.second;

//			s->add_doc(70 + get_doc_at(i));
		    }
	    }


/*
function upDate(s, k, i)
// (s, (k, i-1)) is the canonical reference pair for the active point
 { var oldr = root;
   var (endPoint, r) = test_and_split(s, k, i-1, Txt.charAt(i));

   while (!endPoint)
    { r.addTransition(i, infinity, new State());
      if (oldr != root) oldr.sLink = r; // build suffix-link active-path

      oldr = r;
      var (s,k) = canonize(s.sLink, k, i-1)
      (endPoint, r) = test_and_split(s, k, i-1, Txt.charAt(i))
    }

   if(oldr != root) oldr.sLink = s;

   return new pair(s, k);
 }//upDate
*/
	pair<Node*,int> update(Node *s, int k, int i)
	    // (s, (k,i-1)) is the canonical reference pair for the active point;
	    {
		Node			*oldr = root;
		pair<bool, Node*>	res = test_and_split(s, k, i-1, T[i]);
		bool			end_point = res.first;
		Node			*r = res.second;

//		r->add_doc(get_doc_at(k),k);
		int d_k = k;
//		int docid = get_doc_at(k);
//		r->add_doc(docid,k);
//		r->add_doc(100 + get_doc_at(i));

		while (!end_point)
		    {
			// g'(r,(i,8)) = r' where r' is a new state;
			Node		*r1 = new Node(this);
			r->create_transition(i, INT_MAX, r1);
//			r1->add_doc(get_doc_at(i),i);
//			r1->add_doc(get_doc_at(k),k);

			if (oldr != root)
			    {
				oldr->slink = r;
			    }

			oldr = r;
			pair<Node*, int>	res2 = canonize(s->slink, k, i-1);
			s = res2.first;
			k = res2.second;
			d_k++;

			res = test_and_split(s, k, i-1, T[i]);
			end_point = res.first;
			r = res.second;

//			r->add_doc(get_doc_at(k),k);
		    }

		if (oldr != root) oldr->slink = s;

//		r->add_doc(3000 + get_doc_at(k));
//		oldr->add_doc(4000 + get_doc_at(k));

		return pair<Node*,int>(s, k);
	    }


/*
function test_and_split(s, k, p, t)
 { if(k<=p)
    { // find the t_k transition g'(s,(k',p'))=s' from s
      // k1 is k'  p1 is p' in Ukkonen '95
      var ((k1,p1), s1)  = s[Txt.charAt(k)];

      if (t == Txt.charAt(k1 + p - k + 1))
         return new pair(true, s);
      else
       { var r = new State()
         s.addTransition(k1, k1+p-k,   r);     // s---->r---->s1
         r.addTransition(    k1+p-k+1, p1, s1);
         return new pair(false, r)
       }
    }
   else // k > p;  ? is there a t-transition from s ?
      return new pair(s[t] != null, s);
 }//test_and_split
*/

	pair<bool,Node*> test_and_split(Node *s, int k, int p, string t)
	    {
		map<string, transition>::iterator	it;

//*		s->add_doc(00 + get_doc_at(k),k);	// noen feilaktige
//		s->add_doc(1,1);

		if (k <= p)
		    {
			it = s->d.find(T[k]);
			transition	t1 = it->second;

			if (t == T[t1.k + p - k + 1])
			    {
//				s->add_doc(100 + get_doc_at(k),k);	// noen feilaktige
			    return pair<bool,Node*>(true, s);
			    }
//			s->add_doc(200 + get_doc_at(k),k);	// noen feilaktige

			Node	*r = new Node(this);
			set< pair<int,int> >::iterator	s_it=t1.s->docs.begin(), s_stop=t1.s->docs.end();
//*			for (; s_it!=s_stop; ++s_it)
//*			    r->add_doc(2000 + ((*s_it).first%1000), (*s_it).second);

			// remove transition
			s->d.erase(it);

			// g'(s,(k',k'+p-k)) = r
			s->create_transition(t1.k, t1.k+p-k, r);
			// g'(r,(k'+p-k+1,p')) = s'
			r->create_transition(t1.k+p-k+1, t1.p, t1.s);

//			s->add_doc(anta + 100 + get_doc_at(k),t1.k+p-k);
//			r->add_doc(anta + 200 + get_doc_at(k),t1.p);
//			anta+= 1000;

			return pair<bool,Node*>(false, r);
		    }
		else
		    {
//		s->add_doc(300 + get_doc_at(k),k);	// noen feilaktige
			it = s->d.find(t);
			if (it!=s->d.end())
			    {
//				transition	t1 = it->second;
//				t1.s->add_doc(40 + get_doc_at(t1.k));
				return pair<bool,Node*>(true, s);
			    }
			else
			    {
				return pair<bool,Node*>(false, s);
			    }
		    }
	    }


/*
function canonize(s, k, p)    // s--->...
 { if(p < k) return new pair (s, k);

   // find the t_k transition g'(s,(k',p'))=s' from s
   // k1 is k',  p1 is p' in Ukk' '95
   var ((k1,p1), s1) = s[Txt.charAt(k)];     // s--(k1,p1)-->s1

   while(p1-k1 <= p-k)                       // s--(k1,p1)-->s1--->...
    { k += p1 - k1 + 1;  // remove |(k1,p1)| chars from front of (k,p)
      s = s1;
      if(k <= p)
       { ((k1,p1), s1) = s[Txt.charAt(k)];   // s--(k1,p1)-->s1
       }
    }
   return new pair(s, k);
 }//canonize
*/

	pair<Node*, int> canonize(Node *s, int k, int p)
	    {
		if (p < k)
		    {
			return pair<Node*, int>(s, k);
		    }

//		s->add_doc(2,1);
//		s->add_doc(get_doc_at(k-1), k-1);

		map<string, transition>::iterator	it = s->d.find(T[k]);

//		assert(it != s->d.end());

		transition	t1 = it->second;

		while (t1.p - t1.k <= p - k)
		    {
			k = k + t1.p - t1.k + 1;
			s = t1.s;

//			s->add_doc(3100 + get_doc_at(k),k);

//			s->add_doc(get_doc_at(k-1), k-1);
/*
			cout << ":" << get_doc_at(k);
			for (int x=k-5; x<T.size() && x<k+5; x++)
			    {
				cout << " ";
				if (x==k) cout << "|";
				cout << T[x];
			    }
			cout << endl;
*/
			if (k <= p)
			    {
				it = s->d.find(T[k]);
//				assert( it != s->d.end() );
				t1 = it->second;
			    }

//			t1.s->add_doc(1000 + get_doc_at(k));
		    }

//		t1.s->add_doc(2000 + get_doc_at(k));

		return pair<Node*, int>(s, k);
	    }


        int get_doc_at(int k)
	    {
	        int		i;
	        for (i=0; i<pos_doc.size() && k>pos_doc[i]; i++);
		return real_docid[i];
	    }


	multimap<double, transition>	B;
	Node		*root, *bottom;
	vector<int>	real_docid, pos_doc;
	double		threshold;
};





/*
long long binomial(int n, int m)
{
    long long	bc[n+1][n+1];
    long long	result;

    for (int i=0; i<=n; i++) bc[i][0] = 1;

    for (int j=0; j<=n; j++) bc[j][j] = 1;


    for (int i=1; i<=n; i++)
	for (int j=1; j<i; j++)
	    bc[i][j] = bc[i-1][j-1] + bc[i-1][j];

    result = bc[n][m];

    return result;
}
*/

long long binomial(int n, int m)
{
    long long	**bc;//[n+1][n+1];
    long long	result;

    bc = new long long *[n+1];
    for (int i=0; i<=n; i++)
	bc[i] = new long long[n+1];

    for (int i=0; i<=n; i++) bc[i][0] = 1;

    for (int j=0; j<=n; j++) bc[j][j] = 1;


    for (int i=1; i<=n; i++)
	for (int j=1; j<i; j++)
	    bc[i][j] = bc[i-1][j-1] + bc[i-1][j];

    result = bc[n][m];

    for (int i=0; i<=n; i++)
	delete bc[i];
    delete bc;

    return result;
}





int main(int argc, char *argv[])
{
#ifdef CLIENT
    if (argc!=2)
	{
	    cerr << "Usage: " << argv[0] << " <hostname>" << endl;
	    exit(-1);
	}
#else
    if (argc!=3 && argc!=4)
	{
	    cerr << "Usage: " << argv[0] << " <query.snippetX> <output> [<threshold>]" << endl;
	    exit(-1);
	}

    ifstream	docs(argv[1]);

    if (!docs)
	{
	    cerr << "Error opening " << argv[1] << endl;
	    exit(-1);
	}
#endif	// CLIENT

#ifdef THRESHOLD
    double	threshold = 10;
    if (argc==4) threshold = atoi(argv[3]);
#endif

#ifdef SERVER
    Server	*server = new Server(16);
#endif

    struct timeb	time1;
    ftime(&time1);
//    time_t	time1 = time(NULL);

/*
    SuffixTree	*ST = new SuffixTree();
    vector<string>	s1, s2, s3, s4, s5, s6, s7;

    s1.push_back("cat");
    s1.push_back("ate");
    s1.push_back("cheese");

    s2.push_back("mouse");
    s2.push_back("ate");
    s2.push_back("cheese");
    s2.push_back("too");

    s3.push_back("cat");
    s3.push_back("ate");
    s3.push_back("mouse");
    s3.push_back("too");

    s4.push_back("mice");
    s4.push_back("too");

    s5.push_back("mice");
    s5.push_back("ate");
    s5.push_back("mouse");
    s5.push_back("too");

    s6.push_back("ate");
    s6.push_back("mouse");
    s6.push_back("too");

    s7.push_back("cat");
    s7.push_back("ate");
    s7.push_back("mouse");
    s7.push_back("also");
*/
/*
    s1.push_back("$");
    s1.insert(s1.end(), s2.begin(), s2.end());
    s1.push_back("$");
    s1.insert(s1.end(), s3.begin(), s3.end());
    s1.push_back("$");
    s1.insert(s1.end(), s4.begin(), s4.end());
    s1.push_back("$");
    ST->algorithm2(s1, 1);
*/
//    ST->print();
//    cout << "-----" << endl;
//    ST->algorithm2(s2, 2);
//    ST->print();
//    cout << "-----" << endl;
//    ST->algorithm2(s3, 3);
//    ST->print();
//    cout << "-----" << endl;
//    ST->algorithm2(s4, 4);
/*
    int		pos = 0;
    cout << "0," << pos << ": " << print_strings(s1) << endl;
    pos+= s1.size() +1;
    cout << "1," << pos << ": " << print_strings(s2) << endl;
    pos+= s2.size() +1;
    cout << "2," << pos << ": " << print_strings(s3) << endl;
    pos+= s3.size() +1;
    cout << "3," << pos << ": " << print_strings(s4) << endl;
    pos+= s4.size() +1;
    cout << "4," << pos << ": " << print_strings(s5) << endl;
    pos+= s5.size() +1;
    cout << "5," << pos << ": " << print_strings(s6) << endl;
*/
/*
    ST->addDocument(s1);
    ST->addDocument(s2);
    ST->addDocument(s3);
    ST->addDocument(s4);
    ST->addDocument(s5);
    ST->addDocument(s6);
    ST->addDocument(s7);
    ST->makeGST();

//    cluster_result		Result = ST->baseClusters(set<string>());

    ST->print();
*/
/*
    s1.push_back("m");
    s1.push_back("i");
    s1.push_back("s");
    s1.push_back("s");
    s1.push_back("i");
    s1.push_back("s");
    s1.push_back("s");
    s1.push_back("i");
    s1.push_back("p");
    s1.push_back("p");
    s1.push_back("i");

    ST->algorithm2(s1, 1);
    ST->print();
*/

/*
    query_array		qa;
    char		*query = "mars planet";
    get_query(query, strlen(query), &qa);

    io_init();

    vector<int>		result;
    string		_dtemp;
    int			d;
    SuffixTree		*ST = new SuffixTree();
    map<string,int>	S;
    int			num_docs = 0;

    while (docs >> _dtemp >> d)
	{
	    num_docs++;
	    result.push_back(d);

	    pair<vector<string>, char*>	D;

	    filter_washing = true;
	    D = getdocument(d, false);

	    char	*snippet;
	    generate_snippet(qa, D.second, strlen(D.second), &snippet, "", "", 320);

	    vector<string>	V;

	    for (int k=0, p=0; snippet[k]!='\0' && k<320; k++)
		{
		    while (snippet[k]!='\0' && !isalpha(snippet[k])) k++;
		    p = k;
		    while (snippet[p]!='\0' && isalpha(snippet[p])) p++;

		    if (p>k+1)
			{
			    char	buf[1024];
			    strncpy(buf, &(snippet[k]), p-k);
			    buf[p-k] = '\0';
			    convert_to_lowercase((unsigned char*)buf);
			    cout << buf << " ";
			    string	s(buf);
			    V.push_back(s);

			    map<string,int>::iterator	it=S.find(s);
			    if (it == S.end())
				{
				    S.insert(pair<string,int>(s,1));
				}
			    else
				{
				    (*it).second++;
				}
			}

		    k = p;
		}

	    cout << endl;

	    // (#dokument, #setning, #ord)
**
	    for (int i=0; i<V.size(); i++)
		if (V[i].size() > 0)
		    {
//			ST->algorithm2(V[i], d);
			D.insert(D.end(), V[i].begin(), V[i].end());
		    }
**
	    ST->addDocument(V);
	}

    cout << endl;
    // stopwords:
    map<string,int>::iterator	it=S.begin(), stop=S.end(), old_it;
    int		size = num_docs * 0.4;
    cout << size << endl;
    for (; it!=stop;)
	{
	    old_it = it;
	    ++it;

	    if ((*old_it).second <= size)
		{
		    S.erase(old_it);
		}
	    else
		cout << " " << (*old_it).first;
	}
    cout << endl << endl;
*/

#ifdef CLIENT

    SuffixTree		*ST = new SuffixTree();
    Client		*client = new Client(argv[1]);
    vector<string>	*V = new vector<string>();
    vector<double>	*tfidf = new vector<double>();
//    set<string>		stopwords;
    int			docid;

#ifdef THRESHOLD
    threshold = client->getThreshold();
    ST->setThreshold(threshold);
#endif

    while (client->getDocument(&docid, V, tfidf))
	{
	    ST->addDocument(docid, *V, *tfidf);
	    V->clear();
	    tfidf->clear();
	}
/*
    client->getDocument(&docid, V);
    cout << "stopwords(" << V->size() << "):";
    for (int i=0; i<V->size(); i++)
	{
	stopwords.insert((*V)[i]);
	cout << " " << (*V)[i];
	}
    cout << endl;
*/
    cout << "makeGST()..." << endl;
    ST->makeGST();
    cout << "done" << endl;

    cluster_result		Result = ST->baseClusters();

/*
typedef struct
{
    int		k, p;
    double	score;
} basecluster;

typedef struct
{
    set< pair<int,int> >	docs;
//    vector< pair<int,int> >	P;
    map<int, basecluster>	P;
} single_cluster;


typedef struct
{
    int				docs_in_r;
    vector<single_cluster>	R;
} cluster_result;
*/
//    cout << Result.docs_in_r << " " << Result.R.size() << endl;
    vector< pair< pair<int,double>, vector<string> > >	BCP;

    cout << Result.R.size() << endl;
    for (int i=0; i<Result.R.size(); i++)
	{
	    single_cluster	sc = Result.R[i];
	    multimap<int, basecluster>::iterator	m_it = sc.P.begin();
	    basecluster		bc = (*m_it).second;
	    vector<string>	vs;

	    if (i<20) cout << "(" << bc.score << ") \"";
	    for (int j=bc.k; j<=bc.p; j++)
		{
		    if (i<20) if (j>bc.k) cout << " ";
		    if (i<20) cout << ST->T[j];
		    vs.push_back(ST->T[j]);
		}
	    if (i<20) cout << "\"" << endl;

	    if (i<20) cout << "  " << (*m_it).first << " docs in basecluster:";
	    set< pair<int,int> >::iterator	it=sc.docs.begin(), stop=sc.docs.end();
	    for (; it!=stop; ++it)
		{
		    if (i<20) cout << " " << (*it).first;
		}
	    if (i<20) cout << endl;
//	    docs == doc_nr, k

	    BCP.push_back( pair< pair<int,double>, vector<string> >( pair<int,double>(sc.docs.size(),bc.score), vs) );
	}

    client->sendBCPhrases(BCP);
    vector< vector<string> >	*FinalBC = new vector< vector<string> >();
    client->getFinalBC(FinalBC);
    client->sendDoclist( ST->extract(*FinalBC) );
    /* ... */

    delete client;

#else

    SuffixTree		*ST = new SuffixTree();
    int			num_docs;
    int			buf_size = 1048576;
    char		*buf = new char[buf_size];
//    set<string>		stopwords;
    set<string>		unique_categories;

#ifdef SERVER
    int			split = 1;

    split+= server->numConnections();

    int			clientWords[split];
    for (int i=0; i<split; i++)
	clientWords[i] = 0;

#ifdef THRESHOLD
    threshold/= (double)split;
    for (int i=0; i<server->numConnections(); i++)
	server->sendThreshold(i, threshold);
#endif
#endif

#ifdef THRESHOLD
    ST->setThreshold(threshold);
#endif

    docs.getline(buf, buf_size-1);
    num_docs = atoi(buf);

    for (int doc=0; doc<num_docs; doc++)
	{
	    struct doc	d;

	    docs.getline(buf, buf_size-1);	// docid
	    d.docid = atoi(buf);
//	    cout << "docid: " << d.docid << endl;
	    docs.getline(buf, buf_size-1);	// url
	    d.url = strdup(buf);
	    docs.getline(buf, buf_size-1);	// path
	    d.path = strdup(buf);
	    docs.getline(buf, buf_size-1);	// category
	    d.category = strdup(buf);
	    unique_categories.insert(d.category);
	    docs.getline(buf, buf_size-1);	// title
	    d.title = strdup(buf);
	    docs.getline(buf, buf_size-1);	// snippet
	    d.snippet = strdup(buf);
	    docs.getline(buf, buf_size-1);	// words

	    istringstream	is(buf);
	    string		s;
	    double		v;

	    while (is >> s >> v)
		{
		    d.words.push_back(s);
		    d.tfidf.push_back(v);
		}

#ifdef SERVER
	    int		min = 0;
	    int		val = clientWords[0];

	    for (int i=1; i<split; i++)
		if (clientWords[i] < val)
		    {
			min = i;
			val = clientWords[i];
		    }

	    clientWords[min]+= d.words.size();

	    if (min==0)
	        ST->addDocument(doc, d.words, d.tfidf);
	    else
	        server->addDocument(min-1, doc, d.words, d.tfidf);
#else
	    ST->addDocument(doc, d.words, d.tfidf);
#endif
	    Doc.push_back(d);
	}

#ifdef SERVER
    for (int i=0; i<server->numConnections(); i++)
	{
	    server->sendInt(i, -1);
	    server->sendInt(i, -1);
	}

//    vector<string>	client_stopwords;
#endif
/*
    string	s;
    cout << "Stopwords:";
    while (docs >> s)
	{
	    stopwords.insert(s);
#ifdef SERVER
	    client_stopwords.push_back(s);
#endif
	    cout << " " << s;
	}
    cout << endl;

#ifdef SERVER
    for (int i=0; i<server->numConnections(); i++)
	server->addDocument(i, 0, client_stopwords);
#endif
*/
    cout << "makeGST()..." << endl;
    ST->makeGST();
    cout << "done" << endl;

    cluster_result		Result;
    Result = ST->baseClusters();

#ifdef SERVER
//    {
	// Get top 500 base clusters (merge):
	map< vector<string>, pair<int,double> >		MergeM;
        multimap<double, pair< int,vector<string> > >	BigB;

        for (int i=0; i<Result.R.size(); i++)
    	    {
		single_cluster	sc = Result.R[i];
	        multimap<int, basecluster>::iterator	m_it = sc.P.begin();
		basecluster		bc = (*m_it).second;
	        vector<string>	vs;

	    if (i<20 || ST->T[bc.k]=="mars") cout << "(" << bc.score << ") \"";
	    for (int j=bc.k; j<=bc.p; j++)
		{
		    if (i<20 || ST->T[bc.k]=="mars") if (j>bc.k) cout << " ";
		    if (i<20 || ST->T[bc.k]=="mars") cout << ST->T[j];
		    vs.push_back(ST->T[j]);
		}
	    if (i<20 || ST->T[bc.k]=="mars") cout << "\"" << endl;

	    if (i<20 || ST->T[bc.k]=="mars") cout << "  " << (*m_it).first << " docs in basecluster:";
	    set< pair<int,int> >::iterator	it=sc.docs.begin(), stop=sc.docs.end();
	    for (; it!=stop; ++it)
		{
		    if (i<20 || ST->T[bc.k]=="mars") cout << " " << (*it).first;
		}
	    if (i<20 || ST->T[bc.k]=="mars") cout << endl;

		MergeM.insert( pair<vector<string>, pair<int,double> >(vs, pair<int,double>(sc.docs.size(), bc.score)) );
	    }

	for (int i=0; i<server->numConnections(); i++)
	    {
		map<vector<string>, pair<int,double> >	M;
		M = server->getBCPhrases(i);

		map<vector<string>, pair<int,double> >::iterator	it=M.begin(), stop=M.end(), M_it;
		for (; it!=stop; ++it)
		    {
			if ((M_it=MergeM.find((*it).first)) == MergeM.end())
			    {
//				if ((*it).first[0]=="mars") cout << "Inserting(" << (*it).second.first << "," << (*it).second.second << ") " << print_strings((*it).first) << endl;
				MergeM.insert( pair< vector<string>, pair<int,double> >((*it).first, (*it).second) );
			    }
			else
			    {
//				if ((*it).first[0]=="mars") cout << "Merging(" << (*it).second.first << "," << (*it).second.second << ") " << print_strings((*it).first) << endl;
				int		docs = (*M_it).second.first;
				double		score = (*M_it).second.second;
//				if ((*it).first[0]=="mars") cout << "-------(" << (*M_it).second.first << "," << (*M_it).second.second << ") " << print_strings((*M_it).first) << endl;

				score/= docs;
				docs+= (*it).second.first;
				score*= docs;
				(*M_it).second = pair<int,double>(docs,score);
//				if ((*it).first[0]=="mars") cout << "  (docs,score) is now (" << docs << "," << score << ")" << endl;
			    }
		    }
	    }

	map<vector<string>, pair<int,double> >::iterator	M_it=MergeM.begin(), M_stop=MergeM.end();
	for (; M_it!=M_stop; ++M_it)
	    {
		BigB.insert( pair<double, pair< int, vector<string> > >((*M_it).second.second,
		    pair< int, vector<string> >((*M_it).second.first, (*M_it).first)) );
	    }


	cout << "Top clusters (reverse) [" << BigB.size() << "]:" << endl;
        multimap<double, pair< int, vector<string> > >::iterator	B_it=BigB.end(), B_stop=BigB.begin();
	vector< vector<string> >	FinalBC;
	vector<double>			FinalBCScore;
#if !defined(THRESHOLD)
	for (int b=0; B_it!=B_stop && b<500; b++)
#else
	for (int b=0; B_it!=B_stop; b++)
#endif
	    {
		--B_it;
#ifdef THRESHOLD
		if ((*B_it).first < threshold) break;
#endif
		FinalBC.push_back((*B_it).second.second);
		FinalBCScore.push_back((*B_it).first);// / (double)(*B_it).second.first);
/*
		if ((*B_it).first > 40)
		{
		cout << (*B_it).first << ":";
		for (int i=0; i<(*B_it).second.size(); i++)
		    cout << " " << (*B_it).second[i];
		cout << endl;
		}
*/
	    }

	for (int i=0; i<server->numConnections(); i++)
	    server->sendFinalBC(i, FinalBC);

	vector< set<int> >	FinalBCDocs = ST->extract(FinalBC);
	for (int i=0; i<server->numConnections(); i++)
	    {
		vector< set<int> >	clientBCDocs = server->getDoclist(i);
		for (int j=0; j<FinalBCDocs.size(); j++)
		    {
			FinalBCDocs[j].insert(clientBCDocs[j].begin(), clientBCDocs[j].end());
		    }
	    }

	vector<int>		belonging;
	for (int i=0; i<FinalBC.size(); i++)
	    {
/*
		double	score = 1.0;
	        int	size = 0;
	        double	bonus = 0.0;

	        for (int j=0; j<FinalBC[i].size(); j++)
		    {
			if (FinalBCtfidf[i][j]>0) size++;
			if (Doc[].tfidf>0) size++;
			else bonus+= 0.001;
		    }

		if (size==0) score = 0.0;
	        else if (size==1) score = 0.30;
	        else if (size<=6) score = 1.75 + ((double)size)/2;
	        else score = 4.75;

		score+= bonus;

	        score*= ((double)FinalBCDocs[i].size());

		FinalBCScore.push_back(score);
*/
//+ax+		FinalBCScore[i]*= ((double)FinalBCDocs[i].size());
		belonging.push_back(i);
//		cout << score << ":";
//	        for (int j=0; j<FinalBC[i].size(); j++)
//		    cout << " " << FinalBC[i][j];
//		cout << endl;
	    }

	set<int>		clusters;
	cout << "Calculating clusters..." << endl;
	for (int i=0; i<FinalBC.size(); i++)
	    {
		for (int j=i+1; j<FinalBC.size(); j++)
		    {
			set<int>::iterator	this_it=FinalBCDocs[i].begin(), this_stop=FinalBCDocs[i].end(),
						s_it=FinalBCDocs[j].begin(), s_stop=FinalBCDocs[j].end();

			int		count = 0;
			for (; this_it!=this_stop && s_it!=s_stop; ++this_it)
			    {
			        while (s_it!=s_stop && (*s_it) < (*this_it)) ++s_it;

			        if ((*s_it) == (*this_it)) count++;
			    }

			if (((count*100)/FinalBCDocs[i].size())>50 &&
			    ((count*100)/FinalBCDocs[j].size())>50)
			    {
				belonging[j] = belonging[i];
				clusters.insert(belonging[i]);
			    }
		    }
	    }

	cout << "Saving clusters..." << endl;
	vector< vector<int> >	FinalClusters;
	vector< set<int> >	FCDocs;

	if (clusters.size() > 0)
	    {
//		Result.docs_in_r = 0;

		set<int>::iterator	it=clusters.begin(), stop=clusters.end();
		for (int x=0; it!=stop; ++it,x++)
		    {
//			single_cluster	SC;
			vector<int>	SC;
			set<int>	Docs;

			for (int i=0; i<belonging.size(); i++)
			    {
				if (belonging[i] == *it)
				    {
					SC.push_back(i);

					for (int j=0; j<FinalBC[i].size(); j++)
					    cout << FinalBC[i][j] << " ";
					cout << endl;
/*
					basecluster	bc;
					bc.k = C[i].k;
					bc.p = C[i].p;
					bc.score = FinalBCScore[i];
					SC.P.insert( pair<int, basecluster>(C[i].s->docs.size(), bc) );
					SC.docs.insert(C[i].s->docs.begin(), C[i].s->docs.end());
*/
					Docs.insert(FinalBCDocs[i].begin(), FinalBCDocs[i].end());
				    }
			    }

//			cout << "---" << endl;
//			Result.docs_in_r+= SC.docs.size();
//			Result.R.push_back(SC);
			FinalClusters.push_back(SC);
			FCDocs.push_back(Docs);
			set<int>::iterator x_it=Docs.begin(), x_stop=Docs.end();
			for (; x_it!=x_stop; ++x_it)
			    cout << *x_it << " ";
			cout << endl;
		    }
	    }
	else
	    {
		cout << "No clusters!" << endl;
	    }

/***/
#if 0
	vector<transition>	C;
	vector<double>		Score;
	vector<int>		belonging;
	set<int>		clusters;
        multimap<double, transition>::iterator	b_it=BigB.end(), b_stop=BigB.begin();
	int			b;

        cout << BigB.size() << " baseclusters after merging." << endl;
	for (b=0; b_it!=b_stop && b<500; b++)
	    {
		--b_it;
	        C.push_back((*it).second);
		Score.push_back((*it).first);	// ??
		belonging.push_back(b);		//??
	    }

	/***/
	for (int i=0; i<server->numConnections(); i++)
//	    server->sendFinalBC(i, ??);

	for (int i=0; i<server->numConnections(); i++)
//	    server->getFinalBC(i, ??);

	for (b=0; b_it!=b_stop && b<500; b++)
	    {
		--b_it;
	        C.push_back((*it).second);
		Score.push_back((*it).first);	// ??
		belonging.push_back(b);		//??
	    }
	/***/

	cout << "Calculating clusters..." << endl;
	for (int i=0; i<C.size(); i++)
	    {
		for (int j=i+1; j<C.size(); j++)
		    {
			int		count = C[i].s->compare(C[j].s);

			if (((count*100)/C[i].s->docs.size())>50 &&
			    ((count*100)/C[j].s->docs.size())>50)
			    {
				belonging[j] = belonging[i];
				clusters.insert(belonging[i]);
			    }
		    }
	    }

	cout << "Saving clusters..." << endl;
	if (clusters.size() > 0)
	    {
		Result.docs_in_r = 0;

		set<int>::iterator	it=clusters.begin(), stop=clusters.end();
		for (int x=0; it!=stop; ++it,x++)
		    {
			single_cluster	SC;

			for (int i=0; i<belonging.size(); i++)
			    {
				if (belonging[i] == *it)
				    {
					basecluster	bc;
					bc.k = C[i].k;
					bc.p = C[i].p;
					bc.score = Score[i];
					SC.P.insert( pair<int, basecluster>(C[i].s->docs.size(), bc) );
					SC.docs.insert(C[i].s->docs.begin(), C[i].s->docs.end());
				    }
			    }

			Result.docs_in_r+= SC.docs.size();
			Result.R.push_back(SC);
		    }
	    }
	else
	    {
		cout << "No clusters!" << endl;
	    }
#endif // 0
//    }
/******/
/*
	cluster_result baseClusters(set<string> S)
	    {
		cluster_result		Result;

		B.clear();
		cout << "Calculating base clusters..."; cout.flush();
		root->traverse(0, -1, S);
		cout << "done" << endl;

		vector<transition>	C;
		vector<double>		Score;
		vector<int>		belonging;
		set<int>		clusters;
		multimap<double, transition>::iterator	it=B.end(), stop=B.begin();
		int			b;

		cout << B.size() << " baseclusters." << endl;
#if defined(CLIENT) || defined(SERVER)
		Result.docs_in_r = 0;
#endif

		for (b=0; it!=stop && b<500; b++)
		    {
			--it;
#if defined(CLIENT) || defined(SERVER)
			single_cluster	SC;
			basecluster	bc;
			bc.k = (*it).second.k;
			bc.p = (*it).second.p;
			bc.score = (*it).first;
			SC.P.insert( pair<int, basecluster>((*it).second.s->docs.size(), bc) );
			SC.docs.insert((*it).second.s->docs.begin(), (*it).second.s->docs.end());

			Result.docs_in_r+= SC.docs.size();
			Result.R.push_back(SC);
#else
			C.push_back((*it).second);
			Score.push_back((*it).first);
			belonging.push_back(b);
#endif
		    }

#if !defined(CLIENT) && !defined(SERVER)
		cout << "Calculating clusters..." << endl;
		for (int i=0; i<C.size(); i++)
		    {
			for (int j=i+1; j<C.size(); j++)
			    {
				int		count = C[i].s->compare(C[j].s);

				if (((count*100)/C[i].s->docs.size())>50 &&
				    ((count*100)/C[j].s->docs.size())>50)
				    {
					belonging[j] = belonging[i];
					clusters.insert(belonging[i]);
				    }
			    }
		    }

		cout << "Saving clusters..." << endl;
		if (clusters.size() > 0)
		    {
			Result.docs_in_r = 0;

			set<int>::iterator	it=clusters.begin(), stop=clusters.end();
			for (int x=0; it!=stop; ++it,x++)
			    {
				single_cluster	SC;

				for (int i=0; i<belonging.size(); i++)
				    {
					if (belonging[i] == *it)
					    {
						basecluster	bc;
						bc.k = C[i].k;
						bc.p = C[i].p;
						bc.score = Score[i];
						SC.P.insert( pair<int, basecluster>(C[i].s->docs.size(), bc) );
						SC.docs.insert(C[i].s->docs.begin(), C[i].s->docs.end());
					    }
				    }

				Result.docs_in_r+= SC.docs.size();
				Result.R.push_back(SC);
			    }
		    }
		else
		    {
			cout << "No clusters!" << endl;
		    }
#endif

		return Result;
*/
/******/
#endif

//    time_t	time2 = time(NULL);
    struct timeb	time2;
    ftime(&time2);
    double		seconds;
    seconds = time2.time - time1.time;
    seconds+= (double)time2.millitm / 1000;
    seconds-= (double)time1.millitm / 1000;
    cout << "Finished in " << seconds << " seconds." << endl;

    double	entropy_first_sum = 0.0,
		entropy_second_sum = 0.0;

#if !defined(SERVER)


#define DUMP
#ifdef DUMP
    char	dump_name[128];
    snprintf(dump_name, 127, "dump.%i", Result.R.size());
    dump_name[127] = '\0';
    ofstream	dump(dump_name);
    dump << "0" << endl;

    for (int i=0; i<Result.R.size(); i++)
	{
	    set< pair<int,int> >::iterator	it=Result.R[i].docs.begin(), stop=Result.R[i].docs.end();
	    for (int num_docs=0; it!=stop; ++it,num_docs++)
		{
		    int		docnr = (*it).first;
		    dump << i << " " << Doc[docnr].docid << endl;
		}
	}
#endif


    ofstream	output(argv[2]);

    output << "\
<?xml version=\"1.0\" encoding=\"utf-8\"?>\
<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\
<html xmlns=\"http://www.w3.org/1999/xhtml\">\
<head>\
    <title>Search: Suffix Tree Clustering</title>\
</head>\
<body style=\"font-size:smaller\">\
    <table style=\"padding:10px\">\
	<tr style=\"vertical-align: top\">\
	    <td style=\"width:300px;\"><div style=\"width:240px; border-style:solid; border-size:1; padding:16px; border-color:#6f9fbf; background:#afdfff; vertical-align:top\">" << endl;

    output << "Found " << Result.docs_in_r << " documents in " << Result.R.size() << " clusters.<br /><ol>" << endl;
    for (int i=0; i<Result.R.size(); i++)
	{
	    output << "<li><a href=\"#cluster" << i+1 << "\">";
	    multimap<int, basecluster>::iterator	c_it=Result.R[i].P.end(), c_stop=Result.R[i].P.begin();
	    --c_it;
	    basecluster		bc_best = (*c_it).second;

	    for (; c_it!=c_stop;)
		{
		    --c_it;
		    if ((*c_it).second.score > bc_best.score)
			bc_best = (*c_it).second;
		}

		    int		k, p;
		    k = bc_best.k;
		    p = bc_best.p;

		    for (int m=k; m<=p && m<k+3; m++)
			{
			    output << ST->T[m] << " ";
			}
	    output << "</a>";
	    output << " (" << Result.R[i].docs.size() << ")";
	    output << "</li>" << endl;
	}
    output << "</ol></div></td>\
	    <td style=\"width:600px\">" << endl;

    int		bk = 0;

    for (int i=0; i<Result.R.size(); i++)
	{
	    output << "<a name=\"cluster" << i+1 << "\">" << endl;

    	    if (bk==0)
	        {
	    	    output << "<div style=\"background: white; padding: 6px\">" << endl;
		    bk = 1;
		}
	    else
	        {
		    output << "<div style=\"background: #dfe0d0; padding: 6px\">" << endl;
		    bk = 0;
		}

	    set<string>	H;

	    output << "<div style=\"color:#444\">" << endl;
	    multimap<int, basecluster>::iterator	c_it=Result.R[i].P.end(), c_stop=Result.R[i].P.begin();
//	    for (int j=0; j<Result.R[i].P.size(); j++)
	    for (int j=0; c_it!=c_stop; j++)
		{
		    --c_it;
		    if (j==0) output << "<span style=\"color:black\">";
		    int		k, p;
		    double	score;
		    k = (*c_it).second.k;
		    p = (*c_it).second.p;
		    score = (*c_it).second.score;
//		    k = Result.R[i].P[j].first;
//		    p = Result.R[i].P[j].second;

		    output << (*c_it).first << ":";
		    for (int m=k; m<=p; m++)
			{
			    H.insert(ST->T[m]);
			    output << " " << ST->T[m];
			}
		    output << " (" << score << ")";
		    if (j==0) output << "</span>";
		    output << "<br />" << endl;
		}
	    output << "</div><br />" << endl;

	    map<string,int>	Category;
	    set< pair<int,int> >::iterator	it=Result.R[i].docs.begin(), stop=Result.R[i].docs.end();
	    for (; it!=stop; ++it)
		{
		    int		docnr = (*it).first;
		    map<string,int>::iterator	m_it = Category.find(Doc[docnr].category);
		    if (m_it==Category.end())
			{
			    Category.insert( pair<string,int>(Doc[docnr].category, 1) );
			}
		    else
			{
			    (*m_it).second++;
			}
		}

	    multimap<int,string>	Cat2;
	    map<string,int>::iterator	m_it=Category.begin(), m_stop=Category.end();
	    double			h_k = Result.R[i].docs.size();
	    for (; m_it!=m_stop; ++m_it)
		{
		    double	h_ck = (*m_it).second;
		    entropy_first_sum+= (h_ck/(double)num_docs) * log(h_ck/h_k);

		    Cat2.insert( pair<int,string>((*m_it).second, (*m_it).first) );
		}

//	    cout << "log( binomial(" << ((int)h_k + unique_categories.size() -1) << ", " << (unique_categories.size() -1) << ") ) = "
//		<< "log(" << binomial((int)h_k + unique_categories.size() -1, unique_categories.size() -1) << ") = "
//		<< log((double)binomial((int)h_k + unique_categories.size() -1, unique_categories.size() -1)) << endl;
	    entropy_second_sum+= log((double)binomial((int)h_k + unique_categories.size() -1, unique_categories.size() -1));

	    output << "<div style=\"color:#a77; font-size:smaller\"><span style=\"color:black\">" << Result.R[i].docs.size()
		<< " documents in cluster</span><table>";
	    map<int,string>::iterator	m2_it=Cat2.end(), m2_stop=Cat2.begin();
	    for (; m2_it!=m2_stop;)
		{
		    --m2_it;
		    output << "<tr>";
		    output << "<td style=\"color:black\">" << (*m2_it).first << "</td>";
		    output << "<td>" << (*m2_it).second << "</td>" << endl;
		    output << "</tr>";
		}
	    output << "</table></div><br />" << endl;

	    it=Result.R[i].docs.begin(), stop=Result.R[i].docs.end();
	    for (int num_docs=0; it!=stop && num_docs<30; ++it,num_docs++)	// Show maximum 30 documents per cluster.
		{
		    int		docnr = (*it).first;
		    output << "<div>" << endl;
		    output << "<span style=\"color:blue\">" << Doc[docnr].title << "</span><br />" << endl;
		    output << "<span style=\"color:#666; font-size:smaller\">" << Doc[docnr].path << ", DocID:" << Doc[docnr].docid << "</span><br />" << endl;

		    if (Doc[docnr].snippet == NULL)
			output << "<span style=\"color:#444\">(empty)</span><br />" << endl;
		    else
			output << Doc[docnr].snippet << "<br />" << endl;
/*
		    output << "----------<br />" << endl;
		    for (int j=0; j<Doc[docnr].words.size(); j++)
			if (H.find(Doc[docnr].words[j]) != H.end())
			    {
				output << "<b>" << Doc[docnr].words[j] << "</b> ";
			    }
			else
			    {
				output << Doc[docnr].words[j] << " ";
			    }
		    output << "<br />" << endl;
*/
		    output << "<span style=\"color:green; font-size:smaller\">" << Doc[docnr].url << "</span><br />";
		    output << "<span style=\"color:#a77; font-size:smaller\">" << Doc[docnr].category << "</span><br />" << endl;
		    output << "</div><br />" << endl;
		}

	    if (it!=stop)
		{
		    output << "<div><span style=\"color:#444\">" << Result.R[i].docs.size() -30
			<< " documents are not shown.</span></div><br />" << endl;
		}

	    output << "</div></a>" << endl;
	}

    output << "</td></tr></table>" << endl;
    output << "</body></html>" << endl;

#else

/*
	vector< vector<string> >	FinalBC;
	vector< set<int> >	FinalBCDocs = ST->extract(FinalBC);
	vector<double>		FinalBCScore;
	vector< vector<int> >	FinalClusters;
	vector< set<int> >	FCDocs
*/

    ofstream	output(argv[2]);

    output << "\
<?xml version=\"1.0\" encoding=\"utf-8\"?>\
<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\
<html xmlns=\"http://www.w3.org/1999/xhtml\">\
<head>\
    <title>Search: Suffix Tree Clustering</title>\
</head>\
<body style=\"font-size:smaller\">\
    <table style=\"padding:10px\">\
	<tr style=\"vertical-align: top\">\
	    <td style=\"width:300px;\"><div style=\"width:240px; border-style:solid; border-size:1; padding:16px; border-color:#6f9fbf; background:#afdfff; vertical-align:top\">" << endl;

    output << "Found " << FinalClusters.size() << " clusters.<br /><ol>" << endl;
    multimap<double, pair<int,int> > TopClusters;
/***/
    for (int i=0; i<FinalClusters.size(); i++)
	{
	    int		best_bc = FinalClusters[i][0];
	    double	best_score = FinalBCScore[best_bc] * FCDocs[i].size();

	    for (int j=1; j<FinalClusters[i].size(); j++)
		if (FinalBCScore[FinalClusters[i][j]] * FCDocs[i].size() > best_score)
		    {
			best_bc = FinalClusters[i][j];
			best_score = FinalBCScore[best_bc] * FCDocs[i].size();
		    }

	    TopClusters.insert( pair<double, pair<int,int> >(best_score, pair<int,int>(i, best_bc)) );
	}
/***/

//    for (int i=0; i<FinalClusters.size(); i++)
    multimap<double, pair<int,int> >::iterator	T_it=TopClusters.end(), T_stop=TopClusters.begin();
    for (; T_it!=T_stop;)
	{
	    --T_it;
	    int		i = (*T_it).second.first;
	    output << "<li><a href=\"#cluster" << i+1 << "\">";

	    int		best_bc = (*T_it).second.second;
	    double	best_score = (*T_it).first;

	    for (int j=0; j<FinalBC[best_bc].size(); j++)
		output << FinalBC[best_bc][j] << " ";
/*
	    multimap<int, basecluster>::iterator	c_it=Result.R[i].P.end(), c_stop=Result.R[i].P.begin();
	    --c_it;
	    basecluster		bc_best = (*c_it).second;

	    for (; c_it!=c_stop;)
		{
		    --c_it;
		    if ((*c_it).second.score > bc_best.score)
			bc_best = (*c_it).second;
		}

		    int		k, p;
		    k = bc_best.k;
		    p = bc_best.p;

		    for (int m=k; m<=p && m<k+3; m++)
			{
			    output << ST->T[m] << " ";
			}
*/
	    output << "</a>";
	    output << " (" << FCDocs[i].size() << ")";
	    output << "</li>" << endl;
	}
    output << "</ol></div></td>\
	    <td style=\"width:600px\">" << endl;

    int		bk = 0;

    for (int i=0; i<FinalClusters.size(); i++)
	{
	    output << "<a name=\"cluster" << i+1 << "\">" << endl;

    	    if (bk==0)
	        {
	    	    output << "<div style=\"background: white; padding: 6px\">" << endl;
		    bk = 1;
		}
	    else
	        {
		    output << "<div style=\"background: #dfe0d0; padding: 6px\">" << endl;
		    bk = 0;
		}

	    set<string>	H;

	    output << "<div style=\"color:#444\">" << endl;
	    for (int j=0; j<FinalClusters[i].size(); j++)
		{
		    int		bc = FinalClusters[i][j];
		    if (j==0) output << "<span style=\"color:black\">";
		    output << FinalBCDocs[bc].size() << ":";
		    for (int m=0; m<FinalBC[bc].size(); m++)
			{
			    H.insert(FinalBC[bc][m]);
			    output << " " << FinalBC[bc][m];
			}
		    output << " (" << FinalBCScore[bc] << ")";
		    if (j==0) output << "</span>";
		    output << "<br />" << endl;
		}
	    output << "</div><br />" << endl;

	    map<string,int>	Category;
	    set<int>::iterator	it=FCDocs[i].begin(), stop=FCDocs[i].end();
	    for (; it!=stop; ++it)
		{
		    int		docnr = (*it);
		    map<string,int>::iterator	m_it = Category.find(Doc[docnr].category);
		    if (m_it==Category.end())
			{
			    Category.insert( pair<string,int>(Doc[docnr].category, 1) );
			}
		    else
			{
			    (*m_it).second++;
			}
		}

	    multimap<int,string>	Cat2;
	    map<string,int>::iterator	m_it=Category.begin(), m_stop=Category.end();
	    double			h_k = FCDocs.size();
	    for (; m_it!=m_stop; ++m_it)
		{
		    double	h_ck = (*m_it).second;
		    if (i<10) entropy_first_sum+= (h_ck/(double)num_docs) * log(h_ck/h_k);

		    Cat2.insert( pair<int,string>((*m_it).second, (*m_it).first) );
		}

//	    cout << "log( binomial(" << ((int)h_k + unique_categories.size() -1) << ", " << (unique_categories.size() -1) << ") ) = "
//		<< "log(" << binomial((int)h_k + unique_categories.size() -1, unique_categories.size() -1) << ") = "
//		<< log((double)binomial((int)h_k + unique_categories.size() -1, unique_categories.size() -1)) << endl;
	    if (i<10) entropy_second_sum+= log((double)binomial((int)h_k + unique_categories.size() -1, unique_categories.size() -1));

	    output << "<div style=\"color:#a77; font-size:smaller\"><span style=\"color:black\">" << FCDocs[i].size()
		<< " documents in cluster</span><table>";
	    map<int,string>::iterator	m2_it=Cat2.end(), m2_stop=Cat2.begin();
	    for (; m2_it!=m2_stop;)
		{
		    --m2_it;
		    output << "<tr>";
		    output << "<td style=\"color:black\">" << (*m2_it).first << "</td>";
		    output << "<td>" << (*m2_it).second << "</td>" << endl;
		    output << "</tr>";
		}
	    output << "</table></div><br />" << endl;

	    it=FCDocs[i].begin(), stop=FCDocs[i].end();
	    for (int num_docs=0; it!=stop && num_docs<30; ++it,num_docs++)	// Show maximum 30 documents per cluster.
		{
		    int		docnr = (*it);
		    output << "<div>" << endl;
		    output << "<span style=\"color:blue\">" << Doc[docnr].title << "</span><br />" << endl;
		    output << "<span style=\"color:#666; font-size:smaller\">" << Doc[docnr].path << ", DocID:" << Doc[docnr].docid << "</span><br />" << endl;

		    if (Doc[docnr].snippet == NULL)
			output << "<span style=\"color:#444\">(empty)</span><br />" << endl;
		    else
			output << Doc[docnr].snippet << "<br />" << endl;
/*
		    output << "----------<br />" << endl;
		    for (int j=0; j<Doc[docnr].words.size(); j++)
			if (H.find(Doc[docnr].words[j]) != H.end())
			    {
				output << "<b>" << Doc[docnr].words[j] << "</b> ";
			    }
			else
			    {
				output << Doc[docnr].words[j] << " ";
			    }
		    output << "<br />" << endl;
*/
		    output << "<span style=\"color:green; font-size:smaller\">" << Doc[docnr].url << "</span><br />";
		    output << "<span style=\"color:#a77; font-size:smaller\">" << Doc[docnr].category << "</span><br />" << endl;
		    output << "</div><br />" << endl;
		}

	    if (it!=stop)
		{
		    output << "<div><span style=\"color:#444\">" << FCDocs[i].size() -30
			<< " documents are not shown.</span></div><br />" << endl;
		}

	    output << "</div></a>" << endl;
	}

    output << "</td></tr></table>" << endl;
    output << "</body></html>" << endl;
#endif	// !defined(SERVER)

#ifdef SERVER
    delete server;
#endif

    double	entropy = - entropy_first_sum + (1/(double)num_docs)*entropy_second_sum;
//    cout << "first: " << entropy_first_sum << endl;
//    cout << "second: " << entropy_second_sum << endl;
    cout << "Entropy: " << entropy << endl;

#endif	// CLIENT
}

/*


-------------------
		output_stream << "\
<?xml version=\"1.0\" encoding=\"utf-8\"?>\
<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\
<html xmlns=\"http://www.w3.org/1999/xhtml\">\
<head>\
    <title>Search: \"";

		for (int i=0; i<qa->n; i++)
		    output_stream << qa->query[i].s[0] << " ";

		output_stream << "\"</title>\
</head>\
<body style=\"font-size:smaller\">\
    <table style=\"padding:10px\">\
	<tr style=\"vertical-align: top\">\
	    <td style=\"width:200px;\"><div style=\"width:140px; border-style:solid; border-size:1; padding:16px; border-color:#6f9fbf; background:#afdfff; vertical-align:top\">" << endl;
	    }

	~Output()
	    {
		if (!output_stream) return;
		output_stream << "</td></tr></table>" << endl;
		output_stream << "</body></html>" << endl;
		output_stream.close();
	    }

	void print_sidebar( int num_documents )
	    {
		if (!output_stream) return;

		output_stream << "Found " << num_documents << " documents.\
	    </div></td>\
	    <td style=\"width:600px\">" << endl;
	    }

	void print_hr()
	    {
		if (!output_stream) return;

		output_stream << "<hr />" << endl;
	    }

	void print_snippets( list<int> L, bool verbose=true )
	    {
		if (!output_stream) return;

		if (bk==0)
		    {
			output_stream << "<div style=\"background: white; padding: 6px\">" << endl;
			bk = 1;
		    }
		else
		    {
			output_stream << "<div style=\"background: #dfe0d0; padding: 6px\">" << endl;
			bk = 0;
		    }

		list<int>::iterator	it=L.begin(), stop=L.end();
		for (int i=0; it!=stop; ++it,i++)
		    {
		        struct id_block	*doc = (struct id_block*)mmap_id;
		        doc+= *it;

			if (verbose)
		    	    cout << *it << " " << doc->url << " " << doc->path << " "
				<< doc->category << " \"" << doc->title << "\"" << endl;

			string	preparsed_fname = "preparsed/";
			preparsed_fname+= doc->path;
			FILE	*preparsed_f = fopen(preparsed_fname.c_str(), "r");

			output_stream << "<div>" << endl;
			output_stream << "<span style=\"color:blue\">" << doc->title << "</span><br />" << endl;

			if (!preparsed_f)
			    {
			        output_stream << "Error reading document.<br />" << endl;
			    }
			else
			    {
				struct stat	fileinfo;
				fstat( fileno(preparsed_f), &fileinfo );

				int		size = fileinfo.st_size;
				if (size>0)
				    {
					char	*buf = (char*)malloc(size);

					for (int j=0; j<size;)
					    j+= fread( &(buf[j]), sizeof(char), size-j, preparsed_f );

					char	*snippet;

					generate_snippet( *qa, buf, size, &snippet, "<strong>", "</strong>", 240 );

					if (snippet!=NULL)
					    {
						output_stream << snippet << "<br />" << endl;
						free(snippet);
					    }
					else
					    {
						output_stream << "<br />" << endl;
					    }

					free(buf);
				    }
				else
				    {
					output_stream << "<br />" << endl;
				    }

				fclose(preparsed_f);
			    }

			output_stream << "<span style=\"color:green; font-size:smaller\">" << doc->url << "</span><br />" << endl;
			output_stream << "</div><br />" << endl;
		    }

		output_stream << "</div>" << endl;
	    }
*/

