
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <iostream>
#include <fstream>
#include <list>
#include <vector>
#include <set>
#include <map>
#include <string>
#include <sstream>

#include "blocks.h"

using namespace std;


string print_strings( vector<string> V )
{
    ostringstream	oss;

    int i;
    for (i=0; i<V.size() && i<10; i++)
	{
	    if (i>0) oss << " ";
	    oss << V[i];
	}

    if (i<V.size()) oss << " ...";

    return oss.str();
}

char	*addr_name[] = {"alpha","beta","gamma","delta","epsilon","zeta","eta","theta","iota","lambda","mu","nu","xi","omicron","phi","rho","sigma","tau","upsilon","omega",
			"anker","gul","kaffe","snelle","kantarell","tralle","techno"};
map<int,char*>	addr_hash;
int	last=0;

string print_addr( void *ptr )
{
    return string("<addr>");
    map<int,char*>::iterator	it=addr_hash.find((int)ptr);

    if (it==addr_hash.end())
	{
	    addr_hash.insert(pair<int,char*>((int)ptr, addr_name[last]));
	    return string(addr_name[last++]);
	}
    else
	{
	    return string((*it).second);
	}
}


class SuffixTree;

class SuffixTree
{
    public:

	class Node;

	class Node
	    {
		public:

		    Node(int doc_nr=-1)
			{
			    slink = NULL;

			    if (doc_nr>=0)
				docs.insert(doc_nr);
			}

		    ~Node()
			{
			}

		    void create_transition(vector<string> TT, Node *s)
			{
			    vector<string>	newTT;
			    newTT.push_back(TT[0]);

			    d[newTT] = pair<vector<string>, Node*>(TT, s);
//			    cout << "transition: d[" << TT[0] << "] = (\"" << print_strings(TT) << "\", " << print_addr((void*)s) << ")" << endl;
			}

		    void create_bottom_transition(vector<string> TT, Node *s)
			{
			    d[TT] = pair<vector<string>, Node*>(vector<string>(), s);
//			    cout << "transition: d[" << print_strings(TT) << "] = (\"" << "\", " << print_addr((void*)s) << ")" << endl;
			}

		    void add_doc(int doc_nr)
			{
			    docs.insert(doc_nr);
			}

		    void print_all(int indent=0)
			{
			    map< vector<string>, pair<vector<string>, Node*> >::iterator	it=d.begin(), stop=d.end();
			    for (; it!=stop; ++it)
				{
				    //(T1, s) = (*it).second;
				    vector<string>	T1 = (*it).second.first;
				    Node*		s = (*it).second.second;

				    for (int i=0; i<indent; i++) cout << " ";
				    cout << "|";

				    for (int i=0; i<T1.size(); i++) cout << " " << T1[i];

				    set<int>::iterator	s_it=s->docs.begin(), s_stop=s->docs.end();
				    for (; s_it!=s_stop; ++s_it) cout << " " << *s_it;
//				    cout << " (" << print_addr(s) << ")";

				    cout << endl;
				    s->print_all(indent + T1.size() + 2);
				}
			}


		    list< pair< vector<string>, set<int> > > traverse()
			{
			    vector<string>	empty;
			    return traverse(empty);
			}


		    list< pair< vector<string>, set<int> > > traverse(vector<string> prefix)
			{
			    list< pair< vector<string>, set<int> > >	E;
			    map< vector<string>, pair<vector<string>, Node*> >::iterator	it=d.begin(), stop=d.end();
			    for (; it!=stop; ++it)
				{
				    vector<string>	T1 = (*it).second.first;
				    Node		*s = (*it).second.second;

				    if (s->docs.size() >= 2)
					{
					    vector<string>	newT1;
					    for (int i=0; i<prefix.size(); i++) newT1.push_back(prefix[i]);
					    for (int i=0; i<T1.size(); i++) newT1.push_back(T1[i]);

					    E.push_back( pair< vector<string>, set<int> >(newT1, s->docs) );
					    // E = E + s.traverse(newT1);
					    list< pair< vector<string>, set<int> > >	X;
					    X = s->traverse(newT1);
					    E.splice( E.end(), X );
					}
				}

			    return E;
			}

		    void print_one()
			{
			    cout << "(  ";
			    map< vector<string>, pair<vector<string>, Node*> >::iterator	it=d.begin(), stop=d.end();
			    for (; it!=stop; ++it)
				{
				    vector<string>	V = (*it).first;
				    for (int i=0; i<V.size(); i++)
					cout << V[i] << " ";

				    cout << " ";
				}
			    cout << ")";
			}

		    map< vector<string>, pair<vector<string>, Node*> >	d;
		    Node	*slink;
		    set<int>	docs;
	    };


	SuffixTree()
	    {
		doc_nr = 0;
		root = new Node();
		bottom = new Node();
	    }

	~SuffixTree()
	    {
	    }


	void algorithm2(vector <string> T, int _doc_nr)
	    {
		doc_nr = _doc_nr;

		//for j in range(len(T)):
		for (int j=0; j<T.size(); j++)
		    {
			// create transition g'(bottom, (j,j)) = root
			vector<string>	newT;
			newT.push_back(T[j]);
//			cout << "        bt1: ";
			bottom->create_bottom_transition(newT, root);
		    }
		// create suffix link f'(root) = bottom
		root->slink = bottom;

		Node	*s = root;
		int	k = 0;
		string	tk = T[k];

		//for i in range(len(T)):
		for (int i=0; i<T.size(); i++)
		    {
			pair<Node*,int>	ret1;
			ret1 = update(s, k, i, tk, T);
			s = ret1.first;
			k = ret1.second;

			vector<string>	newT;
			for (int j=k; j<i+2 && j<T.size(); j++) newT.push_back(T[j]);

			pair<Node*, pair<int,string> >	ret2;
//			ret2 = canonize(s, k, i, tk, T[k:i+2]);
			ret2 = canonize(s, k, i, tk, newT);
			s = ret2.first;
			k = ret2.second.first;
			tk = ret2.second.second;
		    }
	    }


	void print()
	    {
		root->print_all();
	    }

	list< pair< vector<string>, set<int> > > calculate_base_clusters()
	    {
		return root->traverse();
	    }


    private:


	pair<Node*,int> update(Node *s, int k, int i, string tk, vector<string> T)
	    {
//		cout << "update(" << print_addr(s) << ", " << k << ", " << i << ", " << tk << ", \"" << print_strings(T) << "\")" << endl;

		//print "u1:", tk
		// (s, (k, i-1)) is the canonical reference pair for the active point
		Node			*oldr = root;
		pair<bool,Node*>	res = test_and_split(s, i-k, T[i], tk);
//		(end_point, r) = test_and_split(s, i-k, T[i], tk);
		bool			end_point = res.first;
		Node			*r = res.second;

		r->add_doc(doc_nr);
//		cout << "adding " << print_addr(r) << " <- " << doc_nr << endl;

		while (!end_point)
		    {
			// create new transition g'(r, (i,inf)) = r' where r' is a new state
			//r.d[T[i]] = (T[i:len(T)], Node())
			//print ((i,len(T)), 0)
			vector<string>	newT;
			for (int j=i; j<T.size(); j++) newT.push_back(T[j]);
//			cout << "        ct1: ";
			r->create_transition(newT, new Node(doc_nr));

			if (oldr != root)
			    {
				// create new suffix link f'(oldr) = r
				oldr->slink = r;
			    }

			vector<string>	newT2;
			for (int j=k; j<i+1; j++) newT2.push_back(T[j]);

			oldr = r;
//			(s, k, tk)
			pair<Node*, pair<int,string> >	res2 = canonize(s->slink, k, i-1, tk, newT2);
			s = res2.first;
			k = res2.second.first;
			tk = res2.second.second;

//			(end_point, r)
			res = test_and_split(s, i-k, T[i], tk);
			end_point = res.first;
			r = res.second;

			r->add_doc(doc_nr);
//			cout << "adding " << print_addr(r) << " <- " << doc_nr << endl;
		    }
		
		if (oldr != root)
		    {
			// create new suffix link f'(oldr) = s
			oldr->slink = s;
		    }
		
		return pair<Node*,int>(s, k);
	    }


	pair<bool,Node*> test_and_split(Node *s, int l, string t, string tk)	// l=p-k+1
	    {
//		cout << "    test_and_split(" << print_addr(s) << ", " << l << ", " << t << ", " << tk << ")" << endl;

		vector<string>		temp_Tk;
		temp_Tk.push_back(tk);
		map< vector<string>, pair<vector<string>, Node*> >::iterator	it=s->d.find(temp_Tk);

		if (l>0 && it!=s->d.end())
		    {
			// let g'(s, (k', p')) = s' be the T[k]-transition from s
			//((k1, p1), s1) = s.d[T[k]]
			vector<string>	T1 = (*it).second.first;
			Node		*s1 = (*it).second.second;
//			(T1, s1) = s.d[tk]

			// Dette er hva Ukkonen glemte!
			if (l>=T1.size())
			    {
				return pair<bool,Node*>(false, s1);
			    }

			if (t == T1[l])
			    {
				return pair<bool,Node*>(true, s);
			    }
			else
			    {
				// replace the T[k]-transition above by transitions
				// g'(s, (k', k'+p-k)) = r and g'(r,(k'+p-k+1,p')) = s'
				// where r is a new state
				Node	*r = new Node(doc_nr);

				set<int>::iterator	s_it=s1->docs.begin(), s_stop=s1->docs.end();
				for (; s_it!=s_stop; ++s_it)
				    {
//				for doc_nr in s1.docs:
					r->add_doc(*s_it);
//					cout << "adding " << print_addr(r) << " <- " << doc_nr << endl;
				    }

				vector<string>	newT1_1, newT1_2;
				for (int i=0; i<l; i++) newT1_1.push_back(T1[i]);
				for (int i=l; i<T1.size(); i++) newT1_2.push_back(T1[i]);

//				cout << "        sp1: ";
				s->create_transition(newT1_1, r);
//				cout << "        sp2: ";
				r->create_transition(newT1_2, s1);

				return pair<bool,Node*>(false, r);
			    }
		    }
		else
		    {
			vector<string>		temp_T;
			temp_T.push_back(t);
			it=s->d.find(temp_T);

//			bool = t in s.d
			//if (bool):
				//(T1, s1) = s.d[t]
				//s1.add_doc(this.doc_nr)

			bool	has = it!=s->d.end();

			if (has)
			    {
				Node		*s1 = (*it).second.second;
				s1->add_doc(doc_nr);
//				cout << "adding " << print_addr(s1) << " <- " << doc_nr << endl;
			    }
			return pair<bool,Node*>(has, s);
		    }
	    }


	pair<Node*, pair<int,string> > canonize(Node *s, int k, int p, string tk, vector<string> T)
	    {
//		cout << "  canonize(" << print_addr(s) << ", " << k << ", " << p << ", " << tk << ", \"" << print_strings(T) << "\")" << endl;

		if (T.size()-1 <= 0)
		    {
			return pair<Node*, pair<int,string> >(s, pair<int,string>(k, tk) );
		    }
//		else if (tk in s.d)
		else
		    {
			map< vector<string>, pair<vector<string>, Node*> >::iterator	it;
/*
			vector<string>		temp_Tk;
			temp_Tk.push_back(tk);
			map< vector<string>, pair<vector<string>, Node*> >::iterator	it=s->d.find(temp_Tk);
*/
			it = find_string(tk, s->d);

			if (it==s->d.end())
			    {
				return pair<Node*, pair<int,string> >(s, pair<int,string>(k, tk) );
			    }

			// find the T[k]-transition g'(s, (k', p')) = s' from s
			//((k1, p1), s1) = s.d[T[k]]
//			vector<string>	*T1 = &((*it).second.first);
			int		T1_size = (*it).second.first.size();
			Node		*s1 = (*it).second.second;
//			(T1, s1) = s.d[tk]
			int		k_inc = 0;

			while (T1_size <= (int)T.size()-2-k_inc)
			    {
//				cout << (int)T1->size() << " <= " << ((int)T.size() -2 -k_inc) << " ( == " << T.size() << " -2 -" << k_inc << " )" << endl;
				k_inc = k_inc + T1_size + 1;

				s = s1;
				if (k_inc<=(int)T.size()-2 && (it=find_string(T[k_inc], s->d)) != s->d.end() )// and T[k_inc] in s.d)
				    {
					// find the T[k]-transition g'(s, (k', p')) = s' from s
					//((k1, p1), s1) = s.d[T[k]]
//					vector<string>		temp_T;
//					temp_T.push_back(T[k_inc]);
//					map< vector<string>, pair<vector<string>, Node*> >::iterator	it=s->d.find(temp_T);
//					if (has_string(T[k_inc], s->d))
//					if (it!=s->d.end())
//					    {
//					T1 = &((*it).second.first);
					T1_size = (*it).second.first.size();
					s1 = (*it).second.second;
//					    }
//					(T1, s1) = s.d[T[k_inc]]
				    }
			    }
//			cout << "done" << endl;

			return pair<Node*, pair<int,string> >(s, pair<int,string>(k+k_inc, T[k_inc]) );
		    }
/*
		else
		    {
			return pair<Node*, pair<int,string> >(s, k, tk);
		    }
*/
	    }


	map< vector<string>, pair<vector<string>, Node*> >::iterator find_string( string t, map< vector<string>, pair<vector<string>, Node*> > d )
	    {
		vector<string>	T;
		T.push_back(t);
		map< vector<string>, pair<vector<string>, Node*> >::iterator	it = d.find(T);
		return it;
	    }


	int		doc_nr;
	Node		*root, *bottom;
};



int main(int argc, char *argv[])
{
    if (argc!=2)
	{
	    cerr << "Usage: " << argv[0] << " <result_docs>" << endl;
	    exit(-1);
	}

    ifstream	docs(argv[1]);

    if (!docs)
	{
	    cerr << "Error opening " << argv[1] << endl;
	    exit(-1);
	}

/*
    SuffixTree	*ST = new SuffixTree();

    vector<string>	s1, s2, s3, s4;

    s1.push_back("cat");
    s1.push_back("ate");
    s1.push_back("cheese");

    s2.push_back("mouse");
    s2.push_back("ate");
    s2.push_back("cheese");
    s2.push_back("too");

    s3.push_back("cat");
    s3.push_back("ate");
    s3.push_back("mouse");
    s3.push_back("too");

    s4.push_back("mice");
    s4.push_back("too");

    ST->algorithm2(s1, 1);
//    ST->print();
//    cout << "-----" << endl;
    ST->algorithm2(s2, 2);
//    ST->print();
//    cout << "-----" << endl;
    ST->algorithm2(s3, 3);
//    ST->print();
//    cout << "-----" << endl;
    ST->algorithm2(s4, 4);
    ST->print();
*/

    FILE	*f_id = fopen("id", "r");
    int		size_id;
    void	*mmap_id;
    struct stat	inode;

    fstat(fileno(f_id), &inode);
    size_id = inode.st_size;

    mmap_id = mmap(0, size_id, PROT_READ, MAP_PRIVATE, fileno(f_id), 0);


    vector<int>		result;
    string		_dtemp;
    int			d;
    SuffixTree		*ST = new SuffixTree();

    while (docs >> _dtemp >> d)
	{
	    result.push_back(d);

    	    struct id_block	*doc_d = (struct id_block*)mmap_id;
	    doc_d+= d;

    	    cout << d << " " << doc_d->url << " " << doc_d->path << " "
		<< doc_d->category << " \"" << doc_d->title << "\"" << endl;

	    string	wordlist_fname = "wordlist/";
	    wordlist_fname+= doc_d->path;
	    ifstream	wordlist_f(wordlist_fname.c_str());

	    if (!wordlist_f)
	        {
	            cerr << "Error reading wordlist-document." << endl;
		    exit(-1);
	        }

	    string		w;
	    vector<string>	V;

	    while (wordlist_f >> w)
		{
		    V.push_back(w);
		    cout << "."; cout.flush();
		}
	    cout << endl;

	    ST->algorithm2(V, d);
	}
}



